"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CalculateGFinalWithoutSizes = exports.CalculateGFinal = exports.CreateEmptyDagreGraph = exports.GetFullGraph = exports.getDependenciesForStep = exports.GetDependencyStepCountersAndNonExistentRefs = exports.getDependenciesLookup = exports.getEdgesLookup = exports.cleanInvalidStepReferences = exports.getEdgesForStepsSubset = exports.getSuccessorSteps = exports.getPredecessorSteps = exports.getGraphDataForStep = void 0;
const Immer = __importStar(require("immer"));
const Steps = __importStar(require("./Steps"));
const Runner = __importStar(require("./Runner"));
const Logging = __importStar(require("./Logging"));
const Common = __importStar(require("./Common"));
const Dagre = __importStar(require("dagre"));
//given a specific stepCounter and all edges, create a new edges and step lookup to be used for display in a dedicated graph
const getGraphDataForStep = (edges, stepCounter, allSteps) => {
    // grab all predecessor and successor data
    const predecessorSteps = (0, exports.getPredecessorSteps)(edges, stepCounter);
    const successorSteps = (0, exports.getSuccessorSteps)(edges, stepCounter);
    const predecessorEdges = (0, exports.getEdgesForStepsSubset)(edges, predecessorSteps);
    const successorEdges = (0, exports.getEdgesForStepsSubset)(edges, successorSteps);
    // clean steps to remove any potential invalid step references
    const allGraphSteps = Object.assign(Object.assign({}, predecessorSteps), successorSteps);
    const cleanGraphSteps = (0, exports.cleanInvalidStepReferences)(allGraphSteps, allSteps);
    const graphData = {
        edges: Object.assign(Object.assign({}, predecessorEdges), successorEdges),
        viewSteps: cleanGraphSteps
    };
    return graphData;
};
exports.getGraphDataForStep = getGraphDataForStep;
const getPredecessorSteps = (edges, stepCounter) => {
    const steps = { [stepCounter.toString()]: { stepCounter: stepCounter } };
    const getPredecessorStepsInternal = (match) => {
        Steps.forAllSteps(edges, (stepCounter, edgeInfo) => {
            if (edgeInfo.includes(match.toString())) {
                steps[stepCounter] = { stepCounter: stepCounter };
                getPredecessorStepsInternal(stepCounter);
            }
        });
    };
    getPredecessorStepsInternal(stepCounter);
    delete steps[0];
    return steps;
};
exports.getPredecessorSteps = getPredecessorSteps;
const getSuccessorSteps = (edges, stepCounter) => {
    const steps = {};
    const getSuccessorStepsInternal = (match) => {
        var _a;
        steps[match] = { stepCounter: match };
        if (!((_a = edges[match]) === null || _a === void 0 ? void 0 : _a.length))
            return;
        const nextEdges = edges[match];
        nextEdges.forEach(edge => {
            getSuccessorStepsInternal(edge);
        });
    };
    getSuccessorStepsInternal(stepCounter);
    delete steps[0];
    return steps;
};
exports.getSuccessorSteps = getSuccessorSteps;
const getEdgesForStepsSubset = (edges, stepsSubset) => {
    const edgesSubset = {};
    const subsetStepCounters = Object.keys(stepsSubset);
    Object.keys(edges).forEach(edgeStepCounter => {
        if (subsetStepCounters.includes(edgeStepCounter)) {
            const newEdgeValues = edges[edgeStepCounter].filter(stepCounter => {
                return subsetStepCounters.includes(stepCounter);
            });
            edgesSubset[edgeStepCounter] = newEdgeValues;
        }
    });
    return edgesSubset;
};
exports.getEdgesForStepsSubset = getEdgesForStepsSubset;
const cleanInvalidStepReferences = (stepsSubset, allSteps) => {
    const cleanSteps = Immer.produce(stepsSubset, draftState => {
        Object.keys(draftState).forEach(stepCounter => {
            if (!allSteps[stepCounter]) {
                delete draftState[stepCounter];
            }
        });
    });
    return cleanSteps;
};
exports.cleanInvalidStepReferences = cleanInvalidStepReferences;
/*
 * get edges lookup returns a lookup table based on both the steps and the references of the steps
 * corner cases
 * 1) a node doesnt have any edges starting at it - this will be an empty array in the return
 * 2) a node refers to a step that no longer exists in steps - this will also have an entry in the return
 */
const getEdgesLookup = (steps) => {
    const edgesLookupSet = {};
    Steps.forAllSteps(steps, (stepCounter, stepData) => {
        switch (stepData.operation.type) {
            case Runner.EStepTypes.sourceInput:
                break;
            case Runner.EStepTypes.sql:
                const dependencies = (0, exports.getDependenciesForStep)(steps, stepData, null);
                dependencies.forEach(sourceStepCounter => {
                    if (!edgesLookupSet[sourceStepCounter]) {
                        //this covers adding an entry that doesnt exist in steps but is a dependency
                        edgesLookupSet[sourceStepCounter] = new Set();
                    }
                    edgesLookupSet[sourceStepCounter].add(stepCounter);
                });
                break;
            default:
                Common.assert(Logging.LoggingArea.Graph, false, "unknown step type in get edges lookup");
        }
    });
    Steps.forAllSteps(steps, (stepCounter, stepData) => {
        if (!edgesLookupSet[stepCounter]) {
            edgesLookupSet[stepCounter] = new Set();
        }
    });
    let edgesLookup = {}; //sourcestepCounter => [targetstepCounter]
    Object.keys(edgesLookupSet).forEach((s) => {
        edgesLookup[s] = Array.from(edgesLookupSet[s]);
    });
    return edgesLookup;
};
exports.getEdgesLookup = getEdgesLookup;
const getDependenciesLookup = (steps) => {
    let dependenciesLookup = {}; //sourcestepCounter => [targetstepCounter]
    Steps.forAllSteps(steps, (stepCounter, stepData) => {
        switch (stepData.operation.type) {
            case Runner.EStepTypes.sourceInput:
                //no dependencies for sources
                dependenciesLookup[stepCounter] = [];
                break;
            case Runner.EStepTypes.sql:
                dependenciesLookup[stepCounter] = (0, exports.getDependenciesForStep)(steps, stepData, null);
                break;
        }
    });
    return dependenciesLookup;
};
exports.getDependenciesLookup = getDependenciesLookup;
const GetDependencyStepCountersAndNonExistentRefs = (steps, dependencies) => {
    const stepCountersToAdd = [];
    const nonExistentRefs = [];
    const dependenciesNameSet = new Set();
    dependencies.forEach(dependency => {
        const { locationName, nodeName } = dependency;
        dependenciesNameSet.add(`${locationName}*${nodeName}`);
    });
    Steps.forAllSteps(steps, (stepCounter, stepData) => {
        const { name, locationName } = stepData.operation;
        if (dependenciesNameSet.has(`${locationName}*${name}`)) {
            stepCountersToAdd.push(stepCounter);
            dependenciesNameSet.delete(`${locationName}*${name}`);
        }
    });
    const leftOvers = Array.from(dependenciesNameSet);
    leftOvers.forEach(name => {
        const [locationName, nodeName] = name.split('*');
        nonExistentRefs.push({ locationName, nodeName });
    });
    return { stepCountersToAdd, nonExistentRefs };
};
exports.GetDependencyStepCountersAndNonExistentRefs = GetDependencyStepCountersAndNonExistentRefs;
const getDependenciesForStep = (allSteps, stepData, index) => {
    const specifiedIndex = !(index === null);
    let dependencies = [];
    const allDependencies = [];
    if (stepData === null)
        return dependencies;
    //iterate through the stepData.operation.dependencies and add them to allDependencies before converting to stepCounters
    stepData.operation.metadata.sourceMapping.forEach((sourceMap, currIndex) => {
        // control flow to account for generating joins for individual sources of a multisource
        if (specifiedIndex && index === currIndex) { // runs once
            sourceMap.dependencies.forEach(dependency => {
                allDependencies.push(dependency);
            });
        }
        else if (!specifiedIndex) { // runs every time
            sourceMap.dependencies.forEach(dependency => {
                allDependencies.push(dependency);
            });
        }
    });
    /**
     * TODO: CD-3610 - pass edges lookup table into this function and use here in place of GetDependencyStepCountersAndNonExistentRefs to retrieve stepCountersToAdd
     */
    const { stepCountersToAdd } = (0, exports.GetDependencyStepCountersAndNonExistentRefs)(allSteps, allDependencies);
    //check against the nameSet to see the if the current step should be added
    stepCountersToAdd.forEach(dependency => {
        if (!dependencies.includes(dependency) && !!dependency) {
            dependencies.push(dependency);
        }
    });
    return dependencies;
};
exports.getDependenciesForStep = getDependenciesForStep;
const GetFullGraph = (g, steps, getWidthHeight, edgesInfo) => {
    Steps.forAllSteps(steps, (stepCounter, stepData) => {
        let [widthOfNode, heightOfNode] = getWidthHeight(stepCounter);
        g.setNode(stepCounter, {
            label: "",
            width: widthOfNode,
            height: heightOfNode
        });
        return true;
    });
    for (let sourceStep in edgesInfo) {
        for (let targetIndex in edgesInfo[sourceStep]) {
            const targetStep = edgesInfo[sourceStep][targetIndex];
            if (sourceStep in steps && targetStep in steps) {
                g.setEdge(sourceStep, targetStep);
            }
        }
    }
    return g;
};
exports.GetFullGraph = GetFullGraph;
const CreateEmptyDagreGraph = (dagreOptions) => {
    let g = new Dagre.graphlib.Graph({});
    g.setGraph(Object.assign({}, dagreOptions));
    g.setDefaultEdgeLabel(function () {
        return {};
    });
    return g;
};
exports.CreateEmptyDagreGraph = CreateEmptyDagreGraph;
const L = Logging.GetLogger(Logging.LoggingArea.Graph);
/**
 * If I call this with steps = {}, and pass in filled out edgesInfo, then I don't have to worry about unnecessary work
 * @param getWidthHeight
 * @param edgesInfo
 * @param steps
 * @param dagreOptions
 * @returns
 */
const CalculateGFinal = (getWidthHeight, edgesInfo, steps, dagreOptions) => {
    let g = new Dagre.graphlib.Graph({});
    //the dagre library modifies this dagreOptions that gets set. make a copy since this is a prop
    g.setGraph(Object.assign({}, dagreOptions));
    g.setDefaultEdgeLabel(function () {
        return {};
    });
    let gFinal = (0, exports.GetFullGraph)(g, steps, getWidthHeight, edgesInfo);
    Dagre.layout(gFinal); // in use by GraphOld component
    return gFinal;
};
exports.CalculateGFinal = CalculateGFinal;
const CalculateGFinalWithoutSizes = (steps) => {
    const edgesInfo = (0, exports.getEdgesLookup)(steps);
    /*
     * don't pass steps, we don't care about filling out the width/height of the nodes since
     * we're just using gFinal for predecessor and successor lookups
     */
    return (0, exports.CalculateGFinal)((stepCounter) => [1, 1], edgesInfo, steps, {});
};
exports.CalculateGFinalWithoutSizes = CalculateGFinalWithoutSizes;
//# sourceMappingURL=GraphOperations.js.map