"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDefaultStepTypes = exports.GetParsedStepTypesSubset = exports.EditNodeTypeTemplateOrMetadataSpec = exports.UpdateNodeTypeDefaultStorageLocation = exports.ConstructMetadataValueWithDefaultStorageLocation = exports.GenerateNewNodeTypeDetails = exports.CreateNodeTypeAndGetTemplateCounter = exports.ThrowIfStepTypeDoesntExist = exports.FilterNonDataVaultTemplates = exports.ParseStepTypesYAML = exports.GetCustomStepTypes = exports.IsSourceStep = exports.userHasAllDefaultStepTypeDetails = exports.parseStepTypeYAML = exports.ParseStepTypeAndCheckErrorsYAML = exports.getStepTypeMetadataSafe = exports.getStepTypeTemplatesSafe = exports.getStepTypeIsDisabledSafe = exports.getNodeSpecSafe = exports.getStepTypeNameSafe = exports.getNodeSpecYAMLSafe = exports.getStepTypeDataSafe = exports.GetDefaultMaterializationTypeFromNodeSpec = exports.isDefaultStepType = exports.isStepTypeValid = exports.EStepTypeDataVersions = void 0;
const Templates = __importStar(require("./Templates"));
const Logging = __importStar(require("./Logging"));
const Runner = __importStar(require("./Runner"));
const Common = __importStar(require("./Common"));
const RunnerOperations = __importStar(require("./RunnerOperations"));
const StepTypes = __importStar(require("./StepTypes"));
const StepTypesLogger = Logging.GetLogger(Logging.LoggingArea.StepTypes);
var EStepTypeDataVersions;
(function (EStepTypeDataVersions) {
    EStepTypeDataVersions[EStepTypeDataVersions["StepTypeDataV0"] = 0] = "StepTypeDataV0";
})(EStepTypeDataVersions = exports.EStepTypeDataVersions || (exports.EStepTypeDataVersions = {}));
const lookupStepTypeDataVersion = (stepTypeData) => {
    /*
     * for now we only have V0, but this will be where we determine the version
     * as we add new ones in the future
     */
    return EStepTypeDataVersions.StepTypeDataV0;
};
var StepTypeDataLookupKey;
(function (StepTypeDataLookupKey) {
    StepTypeDataLookupKey["nodeSpec"] = "nodeSpec";
    StepTypeDataLookupKey["templates"] = "templates";
    StepTypeDataLookupKey["error"] = "error";
    StepTypeDataLookupKey["name"] = "name";
    StepTypeDataLookupKey["isDisabled"] = "isDisabled";
    StepTypeDataLookupKey["version"] = "version";
    StepTypeDataLookupKey["metadata"] = "metadata";
    StepTypeDataLookupKey["defaultStorageLocation"] = "defaultStorageLocation";
})(StepTypeDataLookupKey || (StepTypeDataLookupKey = {}));
// default value is undefined - null is valid in certain cases that we need to handle
const stepTypeDataVersionLookups = {
    [EStepTypeDataVersions.StepTypeDataV0]: {
        [StepTypeDataLookupKey.nodeSpec]: (stepTypeData) => Common.getValueSafe(stepTypeData, ["metadata", "nodeMetadataSpec"], undefined),
        [StepTypeDataLookupKey.templates]: (stepTypeData) => Common.getValueSafe(stepTypeData, ["metadata", "templates"], undefined),
        [StepTypeDataLookupKey.error]: (stepTypeData) => Common.getValueSafe(stepTypeData, ["metadata", "error"], undefined),
        [StepTypeDataLookupKey.name]: (stepTypeData) => Common.getValueSafe(stepTypeData, ["name"], undefined),
        [StepTypeDataLookupKey.isDisabled]: (stepTypeData) => Common.getValueSafe(stepTypeData, ["isDisabled"], undefined),
        [StepTypeDataLookupKey.version]: (stepTypeData) => Common.getValueSafe(stepTypeData, ["version"], undefined),
        [StepTypeDataLookupKey.metadata]: (stepTypeData) => Common.getValueSafe(stepTypeData, ["metadata"], undefined),
        [StepTypeDataLookupKey.defaultStorageLocation]: (stepTypeData) => Common.getValueSafe(stepTypeData, ["defaultStorageLocation"], undefined),
    }
};
const isStepTypeValid = (stepTypeData) => {
    var _a;
    Common.assert(Logging.LoggingArea.StepTypes, !!stepTypeData, "isStepTypeValid recieved invalid stepTypeData");
    return ((_a = stepTypeData === null || stepTypeData === void 0 ? void 0 : stepTypeData.metadata) === null || _a === void 0 ? void 0 : _a.error) === null;
};
exports.isStepTypeValid = isStepTypeValid;
const isDefaultStepType = (stepType) => {
    return stepType in (0, exports.getDefaultStepTypes)();
};
exports.isDefaultStepType = isDefaultStepType;
/**
 * performs a series of checks:
 * 1) does a config attribute exist
 * 2) does a materialization selector exist inside of the config
 * 3) lastly, does the materialization selector have a default
 *
 * if all 3 checks pass, this returns the default specified in the node config.
 * if any of them fail, we return a default materialization type of table.
 *
 * @param nodeSpec a parsed node spec
 * @returns a default materialization type, if one is specified
 */
const GetDefaultMaterializationTypeFromNodeSpec = (nodeSpec) => {
    // firstly, make sure there is a config. it's valid to not have one
    const nodeSpecConfig = Common.getValueSafe(nodeSpec, ["config"], null);
    if (!nodeSpecConfig) {
        return Templates.materializationTypes.table;
    }
    // check to see if the user included a materialization selector in their config
    const materializationTypeGroup = nodeSpecConfig.find(group => {
        return group.items.some(item => item.type === Templates.FirstClassConfigElement.materializationSelector);
    });
    if (!materializationTypeGroup) {
        return Templates.materializationTypes.table;
    }
    // check to grab the default value
    const materializationSpec = materializationTypeGroup.items.find(item => item.type === Templates.FirstClassConfigElement.materializationSelector);
    if (!materializationSpec.default) {
        return Templates.materializationTypes.table;
    }
    return materializationSpec.default;
};
exports.GetDefaultMaterializationTypeFromNodeSpec = GetDefaultMaterializationTypeFromNodeSpec;
///// STEP TYPE DATA ACCESSORS
const getStepTypeDataSafe = (allStepTypeData, sqlType) => {
    const stepTypeData = allStepTypeData[sqlType];
    // Common.assert(Logging.LoggingArea.StepTypes, stepTypeData !== undefined, `getStepTypeDataSafe error with sqlType: ${sqlType}`)
    return stepTypeData;
};
exports.getStepTypeDataSafe = getStepTypeDataSafe;
///// ACCESSORS SPECIFIC TO YAML STEP TYPES
const getNodeSpecYAMLSafe = (stepTypesParsed, sqlType) => {
    return getAttributeFromStepTypeDataSafe(stepTypesParsed, sqlType, StepTypeDataLookupKey.nodeSpec);
};
exports.getNodeSpecYAMLSafe = getNodeSpecYAMLSafe;
///// ACCESSORS SPECIFIC TO PARSED STEP TYPES
const getStepTypeNameSafe = (allStepTypeData, sqlType) => {
    return getAttributeFromStepTypeDataSafe(allStepTypeData, sqlType, StepTypeDataLookupKey.name);
};
exports.getStepTypeNameSafe = getStepTypeNameSafe;
const getNodeSpecSafe = (allStepTypeData, sqlType) => {
    return getAttributeFromStepTypeDataSafe(allStepTypeData, sqlType, StepTypeDataLookupKey.nodeSpec);
};
exports.getNodeSpecSafe = getNodeSpecSafe;
const getStepTypeIsDisabledSafe = (allStepTypeData, sqlType) => {
    return getAttributeFromStepTypeDataSafe(allStepTypeData, sqlType, StepTypeDataLookupKey.isDisabled);
};
exports.getStepTypeIsDisabledSafe = getStepTypeIsDisabledSafe;
const getStepTypeTemplatesSafe = (allStepTypeData, sqlType) => {
    return getAttributeFromStepTypeDataSafe(allStepTypeData, sqlType, StepTypeDataLookupKey.templates);
};
exports.getStepTypeTemplatesSafe = getStepTypeTemplatesSafe;
const getStepTypeMetadataSafe = (stepTypesParsed, sqlType) => {
    return getAttributeFromStepTypeDataSafe(stepTypesParsed, sqlType, StepTypeDataLookupKey.metadata);
};
exports.getStepTypeMetadataSafe = getStepTypeMetadataSafe;
const getAttributeFromStepTypeDataSafe = (allStepTypeData, sqlType, lookupKey) => {
    const stepTypeData = (0, exports.getStepTypeDataSafe)(allStepTypeData, sqlType);
    const stepTypeDataVersion = lookupStepTypeDataVersion(stepTypeData);
    const value = stepTypeDataVersionLookups[stepTypeDataVersion][lookupKey](stepTypeData);
    if (value === undefined) {
        StepTypesLogger.errorContext(Logging.frontEndGlobalContext, `getAttributeFromStepTypeDataSafe error with sql type: ${sqlType} lookupKey: ${lookupKey}`);
    }
    return value;
};
/**
 * Checks parsed step type for errors since stepTypesYAML does not contain error information
 */
const ParseStepTypeAndCheckErrorsYAML = (nodeMetadataSpec) => {
    let parsedNodeSpec;
    //////// node spec validation	
    try {
        // first, try to parse. this throws an error if it doesn't parse successfully
        parsedNodeSpec = Runner.parseYAMLString(nodeMetadataSpec);
        // validate parsed node typing against the run type
        const runTypeResult = Templates.validateRunType(Templates.NodeSpec, parsedNodeSpec);
        if (!runTypeResult.success) {
            throw new Error(runTypeResult.errorString);
        }
        // config specific checks
        if (parsedNodeSpec.config) {
            // check for duplicate attribute names
            const userDefinedAttributeNames = Runner.getAllAttributeNames(parsedNodeSpec);
            if (Common.arrayHasDuplicates(userDefinedAttributeNames)) {
                throw new Error("All attribute names in the config must be unique");
            }
            // check for any first class attribute names in the user config
            let reservedNames = [];
            const firstClassAttributes = new Set([
                ...Object.values(Runner.EFirstClassAttributesColumn),
                ...Object.values(Runner.EFirstClassAttributesNode)
            ]);
            userDefinedAttributeNames.forEach(attributeName => {
                if (firstClassAttributes.has(attributeName)) {
                    reservedNames.push(attributeName);
                }
            });
            if (reservedNames.length) {
                throw new Error(`Reserved attribute name${reservedNames.length > 1 ? "s" : ""} ${reservedNames.join(", ")} can not be used for custom attributes`);
            }
            // check tables specifically as they can appear valid when not in the initial check
            const allTabularConfigs = [];
            parsedNodeSpec.config.forEach(configGroup => {
                const groupTabularConfigs = configGroup.items.filter(item => {
                    return item.type === Templates.SpecUIType.tabular;
                });
                allTabularConfigs.push(...groupTabularConfigs);
            });
            allTabularConfigs.forEach(tabularConfig => {
                if (!tabularConfig.columns || !tabularConfig.columns.length) {
                    throw new Error("Tabular config must have at least one column");
                }
                else {
                    tabularConfig.columns.forEach(column => {
                        //if column is undefined, it'll be caught in validation below
                        if (!!column && Templates.FirstClassConfigElement[column.type]) {
                            throw new Error("First class components can not be used inside of tabular config");
                        }
                    });
                }
                const validateResult = Templates.validateRunType(Templates.TabularSpec, tabularConfig);
                if (!validateResult.success) {
                    throw new Error(validateResult.errorString);
                }
            });
        }
        return { spec: parsedNodeSpec, errorString: null };
    }
    catch (e) {
        if (e instanceof Error) {
            return { spec: null, errorString: e.toString() };
        }
    }
};
exports.ParseStepTypeAndCheckErrorsYAML = ParseStepTypeAndCheckErrorsYAML;
/**
 * validates and parses a passed YAML step type
 * @param allStepTypesYAML - should be the stepTypesYAML dictionary from redux
 * @param stepType - the specific templateID
 * @param allStepTypesParsed - to reference the previous value if the new YAML update is invalid
 * @returns step type data with a parsed node spec and proper error value
 */
const parseStepTypeYAML = (allStepTypesYAML, stepType) => {
    /*
     * assertions are built into these accessors, if they don't exist an exception will be thrown
     * this is something that should never fail, so it's thrown in this context to ensure the app stops
     */
    const yamlNodeSpec = (0, exports.getNodeSpecYAMLSafe)(allStepTypesYAML, stepType);
    const stepTypeData = (0, exports.getStepTypeDataSafe)(allStepTypesYAML, stepType);
    const parsedNodeSpecResult = (0, exports.ParseStepTypeAndCheckErrorsYAML)(yamlNodeSpec);
    const stepTypeParsed = {
        name: stepTypeData.name,
        isDisabled: stepTypeData.isDisabled,
        version: stepTypeData.version,
        metadata: {
            defaultStorageLocation: stepTypeData.metadata.defaultStorageLocation || null,
            templates: stepTypeData.metadata.templates,
            nodeMetadataSpec: parsedNodeSpecResult.spec,
            error: parsedNodeSpecResult.errorString ?
                { errorString: parsedNodeSpecResult.errorString } :
                null
        }
    };
    // only on package imported entities
    if (stepTypeData.packageInfo) {
        stepTypeParsed["packageInfo"] = stepTypeData.packageInfo;
    }
    return stepTypeParsed;
};
exports.parseStepTypeYAML = parseStepTypeYAML;
const userHasAllDefaultStepTypeDetails = (userStepTypeDetails) => {
    const userStepTypeKeys = Object.keys(userStepTypeDetails);
    const defaultStepTypes = (0, exports.getDefaultStepTypes)();
    return Object.keys(defaultStepTypes).every(defaultStepType => userStepTypeKeys.includes(defaultStepType));
};
exports.userHasAllDefaultStepTypeDetails = userHasAllDefaultStepTypeDetails;
const IsSourceStep = (operation) => {
    return operation.sqlType === Runner.ESQLType.source;
};
exports.IsSourceStep = IsSourceStep;
const GetCustomStepTypes = (stepTypes) => {
    const defaultStepTypes = (0, exports.getDefaultStepTypes)();
    const defaultStepTypeNames = Object.values(defaultStepTypes).map(stepType => stepType.name);
    return Object.keys(stepTypes).reduce((customStepTypes, stepTypeKey) => {
        const thisStepType = stepTypes[stepTypeKey];
        // filter out default node types
        if (!defaultStepTypeNames.includes(thisStepType.name)) {
            customStepTypes[stepTypeKey] = thisStepType;
        }
        return customStepTypes;
    }, {});
};
exports.GetCustomStepTypes = GetCustomStepTypes;
const ParseStepTypesYAML = (stepTypesYAML) => {
    if (!stepTypesYAML)
        return {};
    const stepTypesParsed = {};
    Object.keys(stepTypesYAML).forEach(stepType => {
        stepTypesParsed[stepType] = StepTypes.parseStepTypeYAML(stepTypesYAML, stepType);
    });
    return stepTypesParsed;
};
exports.ParseStepTypesYAML = ParseStepTypesYAML;
/**
 * filters out data vault node types from step types for places that require that in the app (i.e. Add Nodes menu)
 * @param stepTypesParsed stepTypesParsed from state to include custom node types
 * @returns all step types without default data vault node types
 */
const FilterNonDataVaultTemplates = (stepTypesParsed) => {
    const filteredStepTypeDetails = Common.CreateDeepClone(stepTypesParsed);
    delete filteredStepTypeDetails["Link"];
    delete filteredStepTypeDetails["Satellite"];
    delete filteredStepTypeDetails["Hub"];
    return filteredStepTypeDetails;
};
exports.FilterNonDataVaultTemplates = FilterNonDataVaultTemplates;
/**
 * checks if a step type exists - if it doesn't, throw an exception
 * @param allStepTypes
 * @param stepTypeID
 */
const ThrowIfStepTypeDoesntExist = (allStepTypes, stepTypeID) => {
    if (!allStepTypes[stepTypeID]) {
        throw new Error(`Error: Step type with ID ${stepTypeID} does not exist`);
    }
};
exports.ThrowIfStepTypeDoesntExist = ThrowIfStepTypeDoesntExist;
const CreateNodeTypeAndGetTemplateCounter = (addEntity) => {
    const newNodeTypeDetails = (0, exports.GenerateNewNodeTypeDetails)();
    return addEntity(newNodeTypeDetails)
        .then(response => response.newEntityData.templateCounter.toString());
};
exports.CreateNodeTypeAndGetTemplateCounter = CreateNodeTypeAndGetTemplateCounter;
const GenerateNewNodeTypeDetails = () => {
    return {
        name: "New Type",
        version: 1,
        isDisabled: false,
        metadata: {
            defaultStorageLocation: null,
            nodeMetadataSpec: `capitalized: New Type
short: NEW
tagColor: gray
plural: New Types`,
            templates: {
                run: { templateString: "" },
                create: { templateString: "" }
            },
            error: null
        }
    };
};
exports.GenerateNewNodeTypeDetails = GenerateNewNodeTypeDetails;
const ConstructMetadataValueWithDefaultStorageLocation = (stepTypeID, templates, stepTypeData) => {
    const defaultStorageLocation = Common.getValueSafe(templates[stepTypeID].metadata, ["defaultStorageLocation"], null);
    return defaultStorageLocation
        ? Object.assign(Object.assign({}, stepTypeData.metadata), { defaultStorageLocation }) : stepTypeData.metadata;
};
exports.ConstructMetadataValueWithDefaultStorageLocation = ConstructMetadataValueWithDefaultStorageLocation;
const UpdateNodeTypeDefaultStorageLocation = (modifyEntity, nodeTypeID, defaultStorageLocation) => {
    return modifyEntity(nodeTypeID, {
        metadata: { defaultStorageLocation }
    });
};
exports.UpdateNodeTypeDefaultStorageLocation = UpdateNodeTypeDefaultStorageLocation;
const EditNodeTypeTemplateOrMetadataSpec = (operationType, modifyEntity, nodeTypeID, // so this is for custom node types only
newValue) => {
    switch (operationType) {
        case RunnerOperations.OperationTypes.create:
        case RunnerOperations.OperationTypes.run:
            return modifyEntity(nodeTypeID.toString(), {
                metadata: {
                    templates: {
                        [operationType]: {
                            templateString: newValue
                        }
                    }
                }
            });
        case RunnerOperations.OperationTypes.nodeMetadataSpec:
            return modifyEntity(nodeTypeID.toString(), {
                metadata: {
                    nodeMetadataSpec: newValue
                }
            });
        default:
            break;
    }
};
exports.EditNodeTypeTemplateOrMetadataSpec = EditNodeTypeTemplateOrMetadataSpec;
const GetParsedStepTypesSubset = (allStepTypesParsed, enabledOnly, ...stepTypesToExclude) => {
    const stepTypesToReturn = Common.CreateDeepClone(allStepTypesParsed);
    if (!!enabledOnly) {
        const disabledStepTypeNames = Object.keys(stepTypesToReturn).filter(stepTypeName => stepTypesToReturn[stepTypeName].isDisabled);
        disabledStepTypeNames.forEach(stepTypeName => {
            delete stepTypesToReturn[stepTypeName];
        });
    }
    if (!!stepTypesToExclude) {
        stepTypesToExclude.forEach(stepType => {
            delete stepTypesToReturn[stepType];
        });
    }
    return stepTypesToReturn;
};
exports.GetParsedStepTypesSubset = GetParsedStepTypesSubset;
/////// DEFAULT STEP TYPES
const getDefaultStepTypes = () => ({
    [Runner.ESQLType.source]: {
        name: "Source",
        isDisabled: false,
        version: 1,
        metadata: {
            defaultStorageLocation: null,
            nodeMetadataSpec: `capitalized: Source
short: null
tagColor: '#D26000'
plural: Sources`,
            templates: {
                [RunnerOperations.OperationTypes.create]: { templateString: `{{ stage('Validating Source Exists') }}
SELECT 1 FROM {{ ref(node.location.name, node.name) }} LIMIT 0
` },
                [RunnerOperations.OperationTypes.run]: { templateString: `{{ stage('Validating Source Exists') }}
SELECT 1 FROM {{ ref(node.location.name, node.name) }} LIMIT 0
` }
            },
            error: null
        }
    },
    [Runner.ESQLType.view]: {
        name: "View",
        isDisabled: true,
        version: 1,
        metadata: {
            defaultStorageLocation: null,
            nodeMetadataSpec: `capitalized: View
short: V
tagColor: '#C4C4C4'
isDisabled: true
plural: Views

config:
  - groupName: Options
    items: 
    - type: materializationSelector
      options:
        - view
      default: view
      isRequired: true

    - type: toggleButton
      attributeName: selectDistinct
      displayName: Distinct

    - type: multisourceToggle

    - type: overrideSQLToggle

    - displayName: Multi Source Strategy
      attributeName: insertStrategy
      type: dropdownSelector
      default: UNION
      options:
      - "UNION"
      - "UNION ALL" 
      isRequired: true
      enableIf: "{% if node.isMultisource %} true {% else %} false {% endif %}"
`,
            templates: {
                [RunnerOperations.OperationTypes.create]: {
                    templateString: `{% if node.override.create.enabled %}
    
    {{ node.override.create.script }}

{% else %}
    {{ stage('Create View') }}
    CREATE OR REPLACE VIEW {{ ref_no_link(node.location.name, node.name) }}
    (
        {% for col in columns %}
            "{{ col.name }}"
            {%- if col.description | length > 0 %} COMMENT '{{ col.description }}'{% endif %}
            {%- if not loop.last -%}, {% endif %}
        {% endfor %}
    )
    {%- if node.description | length > 0 %} COMMENT = '{{ node.description }}'{% endif %}
    AS
    {% for source in sources %}
        SELECT {% if config.selectDistinct %} DISTINCT {% endif %}
        {% for col in source.columns %}
            {{ get_source_transform(col) }} AS "{{ col.name }}"
            {%- if not loop.last -%}, {% endif %}
        {% endfor %}

        {{ source.join }}

        {% if not loop.last %}
            {% if config.insertStrategy in ['UNION', 'UNION ALL'] %}
                {{ config.insertStrategy }}
            {% else %}
                UNION
            {% endif %}
        {% endif %}
    {% endfor %}
{% endif %}
`
                },
                [RunnerOperations.OperationTypes.run]: { templateString: "" }
            },
            error: null
        },
    },
    [Runner.ESQLType.stage]: {
        name: "Stage",
        isDisabled: false,
        version: 1,
        metadata: {
            defaultStorageLocation: null,
            nodeMetadataSpec: `capitalized: Stage
short: STG
plural: Stages
tagColor: '#2EB67D'

config:
- groupName: Options
  items:
  - type: materializationSelector
    default: table
    options:
    - table
    - view
    isRequired: true

  - type: multisourceToggle
    enableIf: "{% if node.materializationType == 'table' %} true {% else %} false {% endif %}" 

  - type: overrideSQLToggle
    enableIf: "{% if node.materializationType == 'view' %} true {% else %} false {% endif %}"
    
  - displayName: Multi Source Strategy
    attributeName: insertStrategy
    type: dropdownSelector
    default: INSERT
    options:
    - "INSERT"
    - "UNION"
    - "UNION ALL"
    isRequired: true
    enableIf: "{% if node.isMultisource %} true {% else %} false {% endif %}"

  - displayName: Truncate Before
    attributeName: truncateBefore
    type: toggleButton
    default: true

  - displayName: Enable Tests
    attributeName: testsEnabled
    type: toggleButton
    default: true
    
  - displayName: Pre-SQL
    attributeName: preSQL
    type: textBox
    syntax: sql
    isRequired: false

  - displayName: Post-SQL
    attributeName: postSQL
    type: textBox
    syntax: sql
    isRequired: false
`,
            templates: {
                [RunnerOperations.OperationTypes.create]: {
                    templateString: `{% if node.override.create.enabled %}
	
	{{ node.override.create.script }}

{% elif node.materializationType == 'table' %}
	{{ stage('Create Stage Table') }}

	CREATE OR REPLACE TABLE {{ ref_no_link(node.location.name, node.name) }}
	(
		{% for col in columns %}
			"{{ col.name }}" {{ col.dataType }}
			{%- if not col.nullable %} NOT NULL
				{%- if col.defaultValue | length > 0 %} DEFAULT {{ col.defaultValue }}{% endif %}
			{% endif %}
			{%- if col.description | length > 0 %} COMMENT '{{ col.description }}'{% endif %}
			{%- if not loop.last -%}, {% endif %}
		{% endfor %}
	)
	{%- if node.description | length > 0 %} COMMENT = '{{ node.description }}'{% endif %}


{% elif node.materializationType == 'view' %}
    {{ stage('Create Stage View') }}

    CREATE OR REPLACE VIEW {{ ref_no_link(node.location.name, node.name) }}
    (
        {% for col in columns %}
            "{{ col.name }}"
            {%- if col.description | length > 0 %} COMMENT '{{ col.description }}'{% endif %}
            {%- if not loop.last -%}, {% endif %}
        {% endfor %}
    )
    {%- if node.description | length > 0 %} COMMENT = '{{ node.description }}'{% endif %}
    AS
    {% for source in sources %}
        SELECT
        {% for col in source.columns %}
            {{ get_source_transform(col) }} AS "{{ col.name }}"
            {%- if not loop.last -%}, {% endif %}
        {% endfor %}

        {{ source.join }}

        {% if not loop.last %}
            {% if config.insertStrategy in ['UNION', 'UNION ALL'] %}
                {{ config.insertStrategy }}
            {% else %}
                UNION
            {% endif %}
        {% endif %}
    {% endfor %}

{% endif %}
`
                },
                [RunnerOperations.OperationTypes.run]: {
                    templateString: `
    {% for test in node.tests if config.testsEnabled %}
        {% if test.runOrder == 'Before' %}
            {{ test_stage(test.name, test.continueOnFailure) }}
            {{ test.templateString }}
        {% endif %}
    {% endfor %}

{% if node.materializationType == 'table' %}
	{% if config.preSQL %}
		{{ stage('Pre-SQL') }}
		{{ config.preSQL }}
	{% endif %}
	
	
	
		{% if config.truncateBefore %}
	
			{{ stage('Truncate Stage Table') }}
			TRUNCATE IF EXISTS {{ ref_no_link(node.location.name, node.name) }}
	
		{% endif %}
	
	
		{% if config.insertStrategy in ['UNION', 'UNION ALL'] %}
			{{ stage( config.insertStrategy + ' Sources' | string ) }}
			INSERT INTO {{ ref_no_link(node.location.name, node.name) }}
				(
					{% for col in columns %}
						"{{ col.name }}"
						{%- if not loop.last -%},{% endif %}
					{% endfor %}
				)
		{% endif %}
	
	
		{% for source in sources %}
	
			{% if config.insertStrategy == 'INSERT' %}
				{{ stage('Insert ' + source.name | string ) }}
	
				INSERT INTO {{ ref_no_link(node.location.name, node.name) }}
				(
					{% for col in source.columns %}
						"{{ col.name }}"
						{%- if not loop.last -%},{% endif %}
					{% endfor %}
				)
			{% endif %}
	
			SELECT
			{% for col in source.columns %}
                {{ get_source_transform(col) }} AS "{{ col.name }}"
				{%- if not loop.last -%}, {% endif %}
			{% endfor %}
	
			{{ source.join }}
	
			{% if config.insertStrategy in ['UNION', 'UNION ALL'] and not loop.last %}
				{{config.insertStrategy}}
			{% endif %}
	
		{% endfor %}
	
	{% if config.postSQL %}
		{{ stage('Post-SQL') }}
		{{ config.postSQL }}
	{% endif %}
{% endif %}

{% if config.testsEnabled %}
	{% for test in node.tests %}
		{% if test.runOrder == 'After' %}
			{{ test_stage(test.name, test.continueOnFailure) }}
			{{ test.templateString }}
        {% endif %}
	{% endfor %}

	{% for column in columns %}
		{% for test in column.tests %}
			{{ test_stage(column.name + ": " + test.name) }}
			{{ test.templateString }}
		{% endfor %}
	{% endfor %}
{% endif %}
`
                }
            },
            error: null
        }
    },
    [Runner.ESQLType.hub]: {
        name: "Hub",
        isDisabled: false,
        version: 1,
        metadata: {
            defaultStorageLocation: null,
            nodeMetadataSpec: `capitalized: Hub
short: 'HUB'
plural: 'Hubs'

tagColor: '#92712E'

config:
- groupName: Options
  items:
  - type: materializationSelector
    isRequired: true
    options: 
    - table
    default: table

  - displayName: Multi Source Strategy
    attributeName: insertStrategy
    type: dropdownSelector
    default: UNION
    options:
    - "UNION"
    - "UNION ALL"
    isRequired: true
    enableIf: "{% if node.isMultisource %} true {% else %} false {% endif %}"

  - displayName: Enable Tests
    attributeName: testsEnabled
    type: toggleButton
    default: true
    
  - displayName: Pre-SQL
    attributeName: preSQL
    type: textBox
    syntax: sql
    isRequired: false

  - displayName: Post-SQL
    attributeName: postSQL
    type: textBox
    syntax: sql
    isRequired: false

systemColumns:
- displayName: "LOAD_DATE"
  transform: CAST(CURRENT_TIMESTAMP AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemCreateDate

- displayName: "RECORD_SOURCE"
  transform: "''"
  dataType: VARCHAR
  placement: end
  attributeName: isSystemRecordSource`,
            templates: {
                [RunnerOperations.OperationTypes.create]: {
                    templateString: `{% if node.materializationType == 'table' %}
				{{ stage('Create Hub Table') }}
			
				CREATE OR REPLACE TABLE {{ ref_no_link(node.location.name, node.name) }}
				(
					{% for col in columns %}
						"{{ col.name }}" {{ col.dataType }}
						{%- if not col.nullable %} NOT NULL
							{%- if col.defaultValue | length > 0 %} DEFAULT {{ col.defaultValue }}{% endif %}
						{% endif %}
						{%- if col.description | length > 0 %} COMMENT '{{ col.description }}'{% endif %}
						{%- if not loop.last -%}, {% endif %}
					{% endfor %}
				)
				{%- if node.description | length > 0 %} COMMENT = '{{ node.description }}'{% endif %}
			
			{% endif %}`
                },
                [RunnerOperations.OperationTypes.run]: {
                    templateString: `
    {% for test in node.tests if config.testsEnabled %}
        {% if test.runOrder == 'Before' %}
            {{ test_stage(test.name, test.continueOnFailure) }}
            {{ test.templateString }}
        {% endif %}
    {% endfor %}

{% if node.materializationType == 'table' %}
	{% if config.preSQL %}
		{{ stage('Pre-SQL') }}
		{{ config.preSQL }}
	{% endif %}


	{{ stage('Merge Hub') }}
	MERGE INTO {{ ref_no_link(node.location.name, node.name) }} "TGT" USING
	(
		{% for source in sources %}
		SELECT
		{% for col in source.columns %}
            {{ get_source_transform(col) }} AS "{{ col.name }}"
			{%- if not loop.last -%}, {% endif %}
		{% endfor %}

		{{ source.join }}

		{% if not loop.last %}
			{{ config.insertStrategy }}
		{% endif %}
	{% endfor %}
	)
	AS "SRC"
	ON
	{% for col in sources[0].columns if (col.sourceColumns[0] 
                                        and col.sourceColumns[0].column 
                                        and col.sourceColumns[0].column.isHubHash
                                        ) -%}
		{% if not loop.first %}
			AND
		{% endif %}
		"SRC"."{{ col.name }}" = "TGT"."{{ col.name }}"
	{% endfor %}
	WHEN NOT MATCHED THEN
	INSERT
	(
		{% for col in columns %}
			"{{ col.name }}"
			{%- if not loop.last -%}, {% endif %}
		{% endfor %}
	) VALUES
	(
		{% for col in columns %}
			"SRC"."{{ col.name }}"
			{%- if not loop.last -%}, {% endif %}
		{% endfor %}
	)


	{% if config.postSQL %}
		{{ stage('Post-SQL') }}
		{{ config.postSQL }}
	{% endif %}
	
{% endif %}

{% if config.testsEnabled %}
	{% for test in node.tests %}
		{% if test.runOrder == 'After' %}
			{{ test_stage(test.name, test.continueOnFailure) }}
			{{ test.templateString }}
        {% endif %}
	{% endfor %}

	{% for column in columns %}
		{% for test in column.tests %}
			{{ test_stage(column.name + ": " + test.name) }}
			{{ test.templateString }}
		{% endfor %}
	{% endfor %}
{% endif %}
			`
                }
            },
            error: null
        },
    },
    [Runner.ESQLType.link]: {
        name: "Link",
        isDisabled: false,
        version: 1,
        metadata: {
            defaultStorageLocation: null,
            nodeMetadataSpec: `capitalized: Link
short: 'LNK'
plural: 'Links'
tagColor: '#CA2287'

config:
- groupName: Options
  items:
  - type: materializationSelector
    isRequired: true
    default: table
    options:
    - table

  - displayName: Multi Source Strategy
    attributeName: insertStrategy
    type: dropdownSelector
    default: UNION
    options:
    - "UNION"
    - "UNION ALL"
    isRequired: true
    enableIf: "{% if node.isMultisource %} true {% else %} false {% endif %}"

  - displayName: Enable Tests
    attributeName: testsEnabled
    type: toggleButton
    default: true
    
  - displayName: Pre-SQL
    attributeName: preSQL
    type: textBox
    syntax: sql
    isRequired: false

  - displayName: Post-SQL
    attributeName: postSQL
    type: textBox
    syntax: sql
    isRequired: false

systemColumns:
- displayName: "LOAD_DATE"
  transform: CAST(CURRENT_TIMESTAMP AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemCreateDate

- displayName: "RECORD_SOURCE"
  transform: "''"
  dataType: VARCHAR
  placement: end
  attributeName: isSystemRecordSource`,
            templates: {
                [RunnerOperations.OperationTypes.create]: {
                    templateString: `{% if node.materializationType == 'table' %}
				{{ stage('Create Link Table') }}
			
				CREATE OR REPLACE TABLE {{ ref_no_link(node.location.name, node.name) }}
				(
					{% for col in columns %}
						"{{ col.name }}" {{ col.dataType }}
						{%- if not col.nullable %} NOT NULL
							{%- if col.defaultValue | length > 0 %} DEFAULT {{ col.defaultValue }}{% endif %}
						{% endif %}
						{%- if col.description | length > 0 %} COMMENT '{{ col.description }}'{% endif %}
						{%- if not loop.last -%}, {% endif %}
					{% endfor %}
				)
				{%- if node.description | length > 0 %} COMMENT = '{{ node.description }}'{% endif %}
			
			{% endif %}`
                },
                [RunnerOperations.OperationTypes.run]: {
                    templateString: `
    {% for test in node.tests if config.testsEnabled %}
        {% if test.runOrder == 'Before' %}
            {{ test_stage(test.name, test.continueOnFailure) }}
            {{ test.templateString }}
        {% endif %}
    {% endfor %}

{% if node.materializationType == 'table' %}
	{% if config.preSQL %}
		{{ stage('Pre-SQL') }}
		{{ config.preSQL }}
	
	{% endif %}

			
	{{ stage('Merge Link') }}
	MERGE INTO {{ ref_no_link(node.location.name, node.name) }} "TGT" USING
	(
		{% for source in sources %}
		SELECT
		{% for col in source.columns %}
            {{ get_source_transform(col) }} AS "{{ col.name }}"
			{%- if not loop.last -%}, {% endif %}
		{% endfor %}

		{{ source.join }}

		{% if not loop.last %}
			{{ config.insertStrategy }}
		{% endif %}
	{% endfor %}
	)
	AS "SRC"
	ON
	{% for col in sources[0].columns if (col.sourceColumns[0] 
                                        and col.sourceColumns[0].column 
                                        and col.sourceColumns[0].column.isLinkHash
                                        ) -%}
		{% if not loop.first %}
			AND
		{% endif %}
		"SRC"."{{ col.name }}" = "TGT"."{{ col.name }}"
	{% endfor %}
	WHEN NOT MATCHED THEN
	INSERT
	(
		{% for col in columns %}
			"{{ col.name }}"
			{%- if not loop.last -%}, {% endif %}
		{% endfor %}
	) VALUES
	(
		{% for col in columns %}
			"SRC"."{{ col.name }}"
			{%- if not loop.last -%}, {% endif %}
		{% endfor %}
	)

	{% if config.postSQL %}
		{{ stage('Post-SQL') }}
		{{ config.postSQL }}	
	{% endif %}
{% endif %}

{% if config.testsEnabled %}
	{% for test in node.tests %}
		{% if test.runOrder == 'After' %}
			{{ test_stage(test.name, test.continueOnFailure) }}
			{{ test.templateString }}
        {% endif %}
	{% endfor %}

	{% for column in columns %}
		{% for test in column.tests %}
			{{ test_stage(column.name + ": " + test.name) }}
			{{ test.templateString }}
		{% endfor %}
	{% endfor %}
{% endif %}
			`
                }
            },
            error: null
        },
    },
    [Runner.ESQLType.satellite]: {
        name: "Satellite",
        isDisabled: false,
        version: 1,
        metadata: {
            defaultStorageLocation: null,
            nodeMetadataSpec: `capitalized: Satellite
short: 'SAT'
plural: 'Satellites'
tagColor: '#627DE9'

config:
- groupName: Options
  items:
  - type: materializationSelector
    isRequired: true
    default: table
    options:
    - table

  - displayName: Multi Source Strategy
    attributeName: insertStrategy
    type: dropdownSelector
    default: UNION
    options:
    - "UNION"
    - "UNION ALL"
    isRequired: true
    enableIf: "{% if node.isMultisource %} true {% else %} false {% endif %}"

  - displayName: Enable Tests
    attributeName: testsEnabled
    type: toggleButton
    default: true
    
  - displayName: Pre-SQL
    attributeName: preSQL
    type: textBox
    syntax: sql
    isRequired: false

  - displayName: Post-SQL
    attributeName: postSQL
    type: textBox
    syntax: sql
    isRequired: false

systemColumns:
- displayName: "LOAD_DATE"
  transform: CAST(CURRENT_TIMESTAMP AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemCreateDate

- displayName: "RECORD_SOURCE"
  transform: "''"
  dataType: VARCHAR
  placement: end
  attributeName: isSystemRecordSource`,
            templates: {
                [RunnerOperations.OperationTypes.create]: {
                    templateString: `{% if node.materializationType == 'table' %}
				{{ stage('Create Satellite Table') }}
			
				CREATE OR REPLACE TABLE {{ ref_no_link(node.location.name, node.name) }}
				(
					{% for col in columns %}
						"{{ col.name }}" {{ col.dataType }}
						{%- if not col.nullable %} NOT NULL
							{%- if col.defaultValue | length > 0 %} DEFAULT {{ col.defaultValue }}{% endif %}
						{% endif %}
						{%- if col.description | length > 0 %} COMMENT '{{ col.description }}'{% endif %}
						{%- if not loop.last -%}, {% endif %}
					{% endfor %}
				)
				{%- if node.description | length > 0 %} COMMENT = '{{ node.description }}'{% endif %}
			
			{% endif %}`
                },
                [RunnerOperations.OperationTypes.run]: {
                    templateString: `
    {% for test in node.tests if config.testsEnabled %}
        {% if test.runOrder == 'Before' %}
            {{ test_stage(test.name, test.continueOnFailure) }}
            {{ test.templateString }}
        {% endif %}
    {% endfor %}

{% if node.materializationType == 'table' %}
		{% if config.preSQL %}
			{{ stage('Pre-SQL') }}
			{{ config.preSQL }}
		{% endif %}
	
		{{ stage('Merge Satellite') }}
		MERGE INTO {{ ref_no_link(node.location.name, node.name) }} "TGT" USING
		(
			{% for source in sources %}
			SELECT
			{% for col in source.columns %}
                {{ get_source_transform(col) }} AS "{{ col.name }}"
				{%- if not loop.last -%}, {% endif %}
			{% endfor %}
	
			{{ source.join }}
	
			{% if not loop.last %}
				{{ config.insertStrategy }}
			{% endif %}
		{% endfor %}
		)
		AS "SRC"
		ON
		{% for col in sources[0].columns if (col.sourceColumns[0] 
                                            and col.sourceColumns[0].column 
                                            and col.sourceColumns[0].column.isChangeHash
                                            ) -%}
			{% if not loop.first %}
				AND
			{% endif %}
			"SRC"."{{ col.name }}" = "TGT"."{{ col.name }}"
		{% endfor %}
		WHEN NOT MATCHED THEN
		INSERT
		(
			{% for col in columns %}
				"{{ col.name }}"
				{%- if not loop.last -%}, {% endif %}
			{% endfor %}
		) VALUES
		(
			{% for col in columns %}
				"SRC"."{{ col.name }}"
				{%- if not loop.last -%}, {% endif %}
			{% endfor %}
		)
	
	{% if config.postSQL %}
		{{ stage('Post-SQL') }}
		{{ config.postSQL }}	
	{% endif %}
	
{% endif %}

{% if config.testsEnabled %}
	{% for test in node.tests %}
		{% if test.runOrder == 'After' %}
			{{ test_stage(test.name, test.continueOnFailure) }}
			{{ test.templateString }}
        {% endif %}
	{% endfor %}

	{% for column in columns %}
		{% for test in column.tests %}
			{{ test_stage(column.name + ": " + test.name) }}
			{{ test.templateString }}
		{% endfor %}
	{% endfor %}
{% endif %}
			`
                }
            },
            error: null
        },
    },
    [Runner.ESQLType.dimension]: {
        name: "Dimension",
        isDisabled: false,
        version: 1,
        metadata: {
            defaultStorageLocation: null,
            nodeMetadataSpec: `capitalized: Dimension
short: DIM
tagColor: '#1E339A'
plural: Dimensions

config:
- groupName: Options
  items:
  - type: materializationSelector
    isRequired: true
    default: table
    options:
    - table
    - view

  - type: multisourceToggle
    enableIf: "{% if node.materializationType == 'table' %} true {% else %} false {% endif %}"

  - type: businessKeyColumns
    isRequired: true

  - type: changeTrackingColumns
    isRequired: false

  - type: overrideSQLToggle
    enableIf: "{% if node.materializationType == 'view' %} true {% else %} false {% endif %}"

  - displayName: Enable Tests
    attributeName: testsEnabled
    type: toggleButton
    default: true
    
  - displayName: Pre-SQL
    attributeName: preSQL
    type: textBox
    syntax: sql
    isRequired: false

  - displayName: Post-SQL
    attributeName: postSQL
    type: textBox
    syntax: sql
    isRequired: false

systemColumns:
- displayName: '{{NODE_NAME}}_KEY'
  transform: ''
  dataType: NUMBER
  placement: beginning
  attributeName: isSurrogateKey

- displayName: SYSTEM_VERSION
  transform: ''
  dataType: NUMBER
  placement: end
  attributeName: isSystemVersion

- displayName: SYSTEM_CURRENT_FLAG
  transform: ''
  dataType: VARCHAR
  placement: end
  attributeName: isSystemCurrentFlag

- displayName: SYSTEM_START_DATE
  transform: CAST(CURRENT_TIMESTAMP AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemStartDate

- displayName: SYSTEM_END_DATE
  transform: CAST('2999-12-31 00:00:00' AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemEndDate

- displayName: SYSTEM_CREATE_DATE
  transform: CAST(CURRENT_TIMESTAMP AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemCreateDate

- displayName: SYSTEM_UPDATE_DATE
  transform: CAST(CURRENT_TIMESTAMP AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemUpdateDate`,
            templates: {
                [RunnerOperations.OperationTypes.create]: { templateString: `{% if node.materializationType == 'table' %}
	{{ stage('Create Dimension Table') }}

	CREATE OR REPLACE TABLE {{ ref_no_link(node.location.name, node.name) }}
	(
		{% for col in columns %}
			"{{ col.name }}" {{ col.dataType }}
			{% if col.isSurrogateKey %}
\t\t        identity
			{% endif %}
			{%- if not col.nullable %} NOT NULL
				{%- if col.defaultValue | length > 0 %} DEFAULT {{ col.defaultValue }}{% endif %}
			{% endif %}
			{%- if col.description | length > 0 %} COMMENT '{{ col.description }}'{% endif %}
			{%- if not loop.last -%}, {% endif %}
		{% endfor %}
	)
	{%- if node.description | length > 0 %} COMMENT = '{{ node.description }}'{% endif %}


{% elif node.materializationType == 'view' %}
	{{ stage('Create Dimension View') }}

	CREATE OR REPLACE VIEW {{ ref_no_link(node.location.name, node.name) }}
	(
		{% for col in columns %}
			"{{ col.name }}"
			{%- if col.description | length > 0 %} COMMENT '{{ col.description }}'{% endif %}
			{%- if not loop.last -%},{% endif %}
		{% endfor %}
	)
	{%- if node.description | length > 0 %} COMMENT = '{{ node.description }}'{% endif %}
	AS
	{% for source in sources %}

		{% if loop.first %}SELECT {% endif %}

		{% for col in source.columns %}
			{% if col.isSurrogateKey or col.isSystemUpdateDate or col.isSystemCreateDate %}
                NULL
			{% else %}
                {{ get_source_transform(col) }}
			{% endif %}
			AS "{{ col.name }}"
			{%- if not loop.last -%}, {% endif %}
		{% endfor %}
		{{ source.join }}

		{% if not loop.last %} UNION ALL {% endif %}
	{% endfor %}

{% endif %}
`
                },
                [RunnerOperations.OperationTypes.run]: { templateString: `
{% set is_type_2 = columns | selectattr("isChangeTracking") | list | length > 0 %}

    {% for test in node.tests if config.testsEnabled %}
        {% if test.runOrder == 'Before' %}
            {{ test_stage(test.name, test.continueOnFailure) }}
            {{ test.templateString }}
        {% endif %}
    {% endfor %}

{% if node.materializationType == 'table' %}

	{% if config.preSQL %}			
		{{ stage('Pre-SQL') }}
		{{ config.preSQL }}
	{% endif %}
	
    {% if is_type_2 %}

        {% for source in sources %}
            {{ stage('MERGE ' + source.name | string) }}
            MERGE INTO {{ ref_no_link(node.location.name, node.name) }} "TGT"
            USING (
            /* New Rows That Don't Exist */
            SELECT
            {% for col in source.columns if not col.isSurrogateKey %}
                {% if col.isSystemVersion %}
                    1
                {% elif col.isSystemCurrentFlag %}
                    'Y'
                {% else %}
                   {{ get_source_transform(col) }}
                {% endif %}
                AS "{{ col.name }}",
            {% endfor %}
                'INSERT_INITAL_VERSION_ROWS' AS "DML_OPERATION"
            {{ source.join }}
            LEFT JOIN {{ ref_no_link(node.location.name, node.name) }} "DIM" ON
            {% for col in source.columns if col.isBusinessKey -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                    {{ get_source_transform(col) }} = "DIM"."{{ col.name }}"
            {% endfor %}
            WHERE
            {% for col in source.columns if col.isBusinessKey -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                "DIM"."{{ col.name }}" IS NULL
            {% endfor %}
            UNION ALL
            /* New Row Needing To Be Inserted Due To Type-2 Column Changes */
            SELECT
            {% for col in source.columns if not col.isSurrogateKey %}
                {% if col.isSystemVersion %}
                    "DIM"."{{ col.name }}" + 1
                {% elif col.isSystemCurrentFlag %}
                    'Y'
                {% else %}
                   {{ get_source_transform(col) }}
                {% endif %}
                AS "{{ col.name }}",
            {% endfor %}
                'INSERT_NEW_VERSION_ROWS' AS "DML_OPERATION"
            {{ source.join }}
            INNER JOIN {{ ref_no_link(node.location.name, node.name) }} "DIM" ON
            {% for col in source.columns if col.isBusinessKey -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                {{ get_source_transform(col) }} = "DIM"."{{ col.name }}"
            {% endfor %}
            WHERE "DIM"."{{ get_value_by_column_attribute("isSystemCurrentFlag") }}" = 'Y'
            {% for col in source.columns if (col.isChangeTracking == true) %}
                {% if loop.first %}
                    AND (
                {% else %}
                    OR
                {% endif %}
                ( NVL( CAST({{ get_source_transform(col) }} as STRING), '**NULL**') <> NVL( CAST("DIM"."{{ col.name }}" as STRING), '**NULL**') )
                {% if loop.last %}
                    )
                {% endif %}
            {% endfor %}
            UNION ALL
            /* Rows Needing To Be Expired Due To Type-2 Column Changes
            In this case, only two columns are merged (version and end date) */
            SELECT
            {%- for col in source.columns if not col.isSurrogateKey %}
                {% if col.isSystemEndDate %}
                    DATEADD(MILLISECONDS, -1, CAST(CURRENT_TIMESTAMP AS TIMESTAMP))
                {% elif col.isSystemCurrentFlag %}
                    'N'
                {% else %}
                    "DIM"."{{ col.name }}"
                {% endif %}
                AS "{{ col.name }}",
            {% endfor -%}
                'update_expired_version_rows' AS "DML_OPERATION"
            {{ source.join }}
            INNER JOIN {{ ref_no_link(node.location.name, node.name) }} "DIM" ON
            {% for col in source.columns if col.isBusinessKey -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                {{ get_source_transform(col) }} = "DIM"."{{ col.name }}"
            {% endfor %}
            WHERE "DIM"."{{ get_value_by_column_attribute("isSystemCurrentFlag") }}" = 'Y'
            {% for col in source.columns if (col.isChangeTracking == true) %}
                {% if loop.first %}
                    AND (
                {% else %}
                    OR
                {% endif %}
                ( NVL( CAST({{ get_source_transform(col) }} as STRING), '**NULL**') <> NVL( CAST("DIM"."{{ col.name }}" as STRING), '**NULL**') )
                {% if loop.last %}
                    )
                {% endif %}
            {% endfor %}
            UNION ALL
            /* Rows Needing To Be Updated Due To Changes To Non-Type-2 source.columns
            This case merges only when there are changes in non-type-2 column updates, but no changes in type-2 columns*/
            SELECT
            {%- for col in source.columns if not col.isSurrogateKey %}
                {% if col.isSystemVersion or col.isSystemCreateDate or col.isSystemStartDate or col.isSystemEndDate %}
                    "DIM"."{{ col.name }}"
                {% elif col.isSystemCurrentFlag %}
                    'Y'
                {% else %}
                    {{ get_source_transform(col) }}
                {% endif %}
                AS "{{ col.name }}",
            {% endfor -%}
                'UPDATE_NON_TYPE2_ROWS' AS "DML_OPERATION"
            {{ source.join }}
            INNER JOIN {{ ref_no_link(node.location.name, node.name) }} "DIM" ON
            {% for col in source.columns if col.isBusinessKey -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                {{ get_source_transform(col) }} = "DIM"."{{ col.name }}"
            {% endfor %}
            WHERE "DIM"."{{ get_value_by_column_attribute("isSystemCurrentFlag") }}" = 'Y'
            AND (
            {% for col in source.columns if (col.isChangeTracking) -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                {{ get_source_transform(col) }} = "DIM"."{{ col.name }}"
            {% endfor %} )
            {% for col in source.columns if not (   col.isBusinessKey or
                                                    col.isChangeTracking or
                                                    col.isSurrogateKey or
                                                    col.isSystemVersion or
                                                    col.isSystemCurrentFlag or
                                                    col.isSystemStartDate or
                                                    col.isSystemEndDate or
                                                    col.isSystemUpdateDate or
                                                    col.isSystemCreateDate) -%}
                {% if loop.first %}
                    AND (
                {% endif %}
                {% if not loop.first %}
                    OR
                {% endif %}
                NVL( CAST({{ get_source_transform(col) }} as STRING), '**NULL**') <> NVL( CAST("DIM"."{{ col.name }}" as STRING), '**NULL**')
                {% if loop.last %}
                    )
                {% endif %}
            {% endfor %}
        ) AS "SRC"
        ON
        {% for col in source.columns if col.isBusinessKey -%}
            {% if not loop.first %}
                AND
            {% endif %}
            "TGT"."{{ col.name }}" = "SRC"."{{ col.name }}"
        {% endfor %}
        AND "TGT"."{{ get_value_by_column_attribute("isSystemVersion") }}" = "SRC"."{{ get_value_by_column_attribute("isSystemVersion") }}"
        WHEN MATCHED THEN UPDATE SET
        {%- for col in source.columns if not (col.isBusinessKey or col.isSurrogateKey or col.isSystemCreateDate) %}
            "TGT"."{{ col.name }}" = "SRC"."{{ col.name }}"
            {% if not loop.last %}, {% endif %}
        {% endfor -%}
        WHEN NOT MATCHED THEN INSERT (
        {%- for col in source.columns if not col.isSurrogateKey %}
            "{{ col.name }}"
            {% if not loop.last %}, {% endif %}
        {% endfor -%}
        )
        VALUES (
        {%- for col in source.columns if not col.isSurrogateKey %}
            "SRC"."{{ col.name }}"
            {% if not loop.last %}, {% endif %}
        {% endfor -%}
        )

    {% endfor %}



    {% else %}
        {% for source in sources %}
            {{ stage('MERGE ' + source.name | string ) }}
            MERGE INTO {{ ref_no_link(node.location.name, node.name) }} "TGT"
            USING (
                SELECT
                {% for col in source.columns if not col.isSurrogateKey %}
                    {% if col.isSystemVersion %}
                    	1
                    {% elif col.isSystemCurrentFlag %}
                    	'Y'
                    {% else %}
                        {{ get_source_transform(col) }}
                    {% endif %}
                    AS "{{ col.name }}"
                    {%- if not loop.last %}, {% endif %}
                {% endfor %}
                {{ source.join }})
                AS "SRC"
            ON
            {% for col in source.columns if col.isBusinessKey -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                "SRC"."{{ col.name }}" = "TGT"."{{ col.name }}"
            {% endfor %}
            WHEN MATCHED
            {% for col in source.columns if not (   col.isBusinessKey or
                                                    col.isSurrogateKey or
                                                    col.isSystemVersion or
                                                    col.isSystemCurrentFlag or
                                                    col.isSystemStartDate or
                                                    col.isSystemEndDate or
                                                    col.isSystemUpdateDate or
                                                    col.isSystemCreateDate) %}
                {% if loop.first %}
                    AND (
                {% else %}
                    OR
                {% endif %}
                NVL( CAST("SRC"."{{ col.name }}" as STRING), '**NULL**') <> NVL( CAST("TGT"."{{ col.name }}" as STRING), '**NULL**')
                {% if loop.last %}
                    )
                {% endif %}
            {% endfor %}
            THEN UPDATE SET
            {%- for col in source.columns if not (  col.isBusinessKey or
                                                    col.isSurrogateKey or
                                                    col.isSystemVersion or
                                                    col.isSystemCurrentFlag or
                                                    col.isSystemStartDate or
                                                    col.isSystemEndDate or
                                                    col.isSystemCreateDate) %}
                    "TGT"."{{ col.name }}" = "SRC"."{{ col.name }}"
                {% if not loop.last %}, {% endif %}
            {% endfor %}
            WHEN NOT MATCHED THEN
            INSERT (
            {%- for col in source.columns if not col.isSurrogateKey %}
                "{{ col.name }}"
                {% if not loop.last %}, {% endif %}
            {% endfor -%}
            )
            VALUES (
            {%- for col in source.columns if not col.isSurrogateKey %}
                "SRC"."{{ col.name }}"
                {% if not loop.last %}, {% endif %}
            {% endfor -%}
            )
        {% endfor %}
    {% endif %}
	
	{% if config.postSQL %}			
		{{ stage('Post-SQL') }}
		{{ config.postSQL }}
	{% endif %}
{% endif %}

{% if config.testsEnabled %}
	{% for test in node.tests %}
		{% if test.runOrder == 'After' %}
			{{ test_stage(test.name, test.continueOnFailure) }}
			{{ test.templateString }}
        {% endif %}
	{% endfor %}

	{% for column in columns %}
		{% for test in column.tests %}
			{{ test_stage(column.name + ": " + test.name) }}
			{{ test.templateString }}
		{% endfor %}
	{% endfor %}
{% endif %}
` }
            },
            error: null
        },
    },
    [Runner.ESQLType.persistentStage]: {
        name: "Persistent Stage",
        isDisabled: false,
        version: 1,
        metadata: {
            defaultStorageLocation: null,
            nodeMetadataSpec: `capitalized: Persistent Stage
short: PSTG
plural: Persistent Stages
tagColor: '#29B2DB'
    
config:
- groupName: Options
  items:
  - type: materializationSelector
    isRequired: true
    default: table
    options:
    - table
    - view
  
  - type: businessKeyColumns
    isRequired: false

  - type: multisourceToggle
    enableIf: "{% if node.materializationType == 'table' %} true {% else %} false {% endif %}"

  - type: overrideSQLToggle
    enableIf: "{% if node.materializationType == 'view' %} true {% else %} false {% endif %}"

  - displayName: Enable Tests
    attributeName: testsEnabled
    type: toggleButton
    default: true
    
  - displayName: Pre-SQL
    attributeName: preSQL
    type: textBox
    syntax: sql
    isRequired: false

  - displayName: Post-SQL
    attributeName: postSQL
    type: textBox
    syntax: sql
    isRequired: false

systemColumns:

- displayName: '{{NODE_NAME}}_KEY'
  transform: ''
  dataType: NUMBER
  placement: beginning
  attributeName: isSurrogateKey

- displayName: SYSTEM_VERSION
  transform: ''
  dataType: NUMBER
  placement: end
  attributeName: isSystemVersion

- displayName: SYSTEM_CURRENT_FLAG
  transform: ''
  dataType: VARCHAR
  placement: end
  attributeName: isSystemCurrentFlag

- displayName: SYSTEM_START_DATE
  transform: CAST(CURRENT_TIMESTAMP AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemStartDate

- displayName: SYSTEM_END_DATE
  transform: CAST('2999-12-31 00:00:00' AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemEndDate

- displayName: SYSTEM_CREATE_DATE
  transform: CAST(CURRENT_TIMESTAMP AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemCreateDate

- displayName: SYSTEM_UPDATE_DATE
  transform: CAST(CURRENT_TIMESTAMP AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemUpdateDate
`,
            templates: {
                [RunnerOperations.OperationTypes.create]: { templateString: `{% if node.materializationType == 'table' %}
    {{ stage('Create Persistent Stage Table') }}

    CREATE OR REPLACE TABLE {{ ref_no_link(node.location.name, node.name) }}
    (
        {% for col in columns %}
            "{{ col.name }}" {{ col.dataType }}
            {% if col.isSurrogateKey %}
\t\t        identity
            {% endif %}
            {%- if not col.nullable %} NOT NULL
                {%- if col.defaultValue | length > 0 %} DEFAULT {{ col.defaultValue }}{% endif %}
            {% endif %}
            {%- if col.description | length > 0 %} COMMENT '{{ col.description }}'{% endif %}
            {%- if not loop.last -%}, {% endif %}
        {% endfor %}
    )
    {%- if node.description | length > 0 %} COMMENT = '{{ node.description }}'{% endif %}


{% elif node.materializationType == 'view' %}
    {{ stage('Create Persistent Stage View') }}

    CREATE OR REPLACE VIEW {{ ref_no_link(node.location.name, node.name) }}
    (
        {% for col in columns %}
            "{{ col.name }}"
            {%- if col.description | length > 0 %} COMMENT '{{ col.description }}'{% endif %}
            {%- if not loop.last -%},{% endif %}
        {% endfor %}
    )
    {%- if node.description | length > 0 %} COMMENT = '{{ node.description }}'{% endif %}
    AS
    {% for source in sources %}

		{% if not loop.last %} UNION ALL {% endif %}
	{% endfor %}

        {% for col in source.columns %}
            {% if col.isSurrogateKey or col.isSystemUpdateDate or col.isSystemCreateDate %}
                NULL
            {% else %}
                {{ get_source_transform(col) }}
            {% endif %}
            AS "{{ col.name }}"
            {%- if not loop.last -%}, {% endif %}
        {% endfor %}
        {{ source.join }}

{% endif %}
` },
                [RunnerOperations.OperationTypes.run]: { templateString: `
{% set has_business_key = columns | selectattr("isBusinessKey") | list | length > 0 %}
{% set is_type_2 = columns | selectattr("isChangeTracking") | list | length > 0 %}

    {% for test in node.tests if config.testsEnabled %}
        {% if test.runOrder == 'Before' %}
            {{ test_stage(test.name, test.continueOnFailure) }}
            {{ test.templateString }}
        {% endif %}
    {% endfor %}

{% if node.materializationType == 'table' %}
	{% if config.preSQL %}
		{{ stage('Pre-SQL') }}
		{{ config.preSQL }}
	{% endif %}
	
    {% if has_business_key and is_type_2 %}

        {% for source in sources %}
            {{ stage('MERGE ' + source.name | string) }}
            MERGE INTO {{ ref_no_link(node.location.name, node.name) }} "TGT"
            USING (
            /* New Rows That Don't Exist */
            SELECT
            {% for col in source.columns if not col.isSurrogateKey %}
                {% if col.isSystemVersion %}
                    1
                {% elif col.isSystemCurrentFlag %}
                    'Y'
                {% else %}
                   {{ get_source_transform(col) }}
                {% endif %}
                AS "{{ col.name }}",
            {% endfor %}
                'INSERT_INITAL_VERSION_ROWS' AS "DML_OPERATION"
            {{ source.join }}
            LEFT JOIN {{ ref_no_link(node.location.name, node.name) }} "DIM" ON
            {% for col in source.columns if col.isBusinessKey -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                    {{ get_source_transform(col) }} = "DIM"."{{ col.name }}"
            {% endfor %}
            WHERE
            {% for col in source.columns if col.isBusinessKey -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                "DIM"."{{ col.name }}" IS NULL
            {% endfor %}
            UNION ALL
            /* New Row Needing To Be Inserted Due To Type-2 Column Changes */
            SELECT
            {% for col in source.columns if not col.isSurrogateKey %}
                {% if col.isSystemVersion %}
                    "DIM"."{{ col.name }}" + 1
                {% elif col.isSystemCurrentFlag %}
                    'Y'
                {% else %}
                   {{ get_source_transform(col) }}
                {% endif %}
                AS "{{ col.name }}",
            {% endfor %}
                'INSERT_NEW_VERSION_ROWS' AS "DML_OPERATION"
            {{ source.join }}
            INNER JOIN {{ ref_no_link(node.location.name, node.name) }} "DIM" ON
            {% for col in source.columns if col.isBusinessKey -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                {{ get_source_transform(col) }} = "DIM"."{{ col.name }}"
            {% endfor %}
            WHERE "DIM"."{{ get_value_by_column_attribute("isSystemCurrentFlag") }}" = 'Y'
            {% for col in source.columns if (col.isChangeTracking == true) %}
                {% if loop.first %}
                    AND (
                {% else %}
                    OR
                {% endif %}
                ( NVL( CAST({{ get_source_transform(col) }} as STRING), '**NULL**') <> NVL( CAST("DIM"."{{ col.name }}" as STRING), '**NULL**') )
                {% if loop.last %}
                    )
                {% endif %}
            {% endfor %}
            UNION ALL
            /* Rows Needing To Be Expired Due To Type-2 Column Changes
            In this case, only two columns are merged (version and end date) */
            SELECT
            {%- for col in source.columns if not col.isSurrogateKey %}
                {% if col.isSystemEndDate %}
                    DATEADD(MILLISECONDS, -1, CAST(CURRENT_TIMESTAMP AS TIMESTAMP))
                {% elif col.isSystemCurrentFlag %}
                    'N'
                {% else %}
                    "DIM"."{{ col.name }}"
                {% endif %}
                AS "{{ col.name }}",
            {% endfor -%}
                'update_expired_version_rows' AS "DML_OPERATION"
            {{ source.join }}
            INNER JOIN {{ ref_no_link(node.location.name, node.name) }} "DIM" ON
            {% for col in source.columns if col.isBusinessKey -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                {{ get_source_transform(col) }} = "DIM"."{{ col.name }}"
            {% endfor %}
            WHERE "DIM"."{{ get_value_by_column_attribute("isSystemCurrentFlag") }}" = 'Y'
            {% for col in source.columns if (col.isChangeTracking == true) %}
                {% if loop.first %}
                    AND (
                {% else %}
                    OR
                {% endif %}
                ( NVL( CAST({{ get_source_transform(col) }} as STRING), '**NULL**') <> NVL( CAST("DIM"."{{ col.name }}" as STRING), '**NULL**') )
                {% if loop.last %}
                    )
                {% endif %}
            {% endfor %}
            UNION ALL
            /* Rows Needing To Be Updated Due To Changes To Non-Type-2 source.columns
            This case merges only when there are changes in non-type-2 column updates, but no changes in type-2 columns*/
            SELECT
            {%- for col in source.columns if not col.isSurrogateKey %}
                {% if col.isSystemVersion or col.isSystemCreateDate or col.isSystemStartDate or col.isSystemEndDate %}
                    "DIM"."{{ col.name }}"
                {% elif col.isSystemCurrentFlag %}
                    'Y'
                {% else %}
                    {{ get_source_transform(col) }}
                {% endif %}
                AS "{{ col.name }}",
            {% endfor -%}
                'UPDATE_NON_TYPE2_ROWS' AS "DML_OPERATION"
            {{ source.join }}
            INNER JOIN {{ ref_no_link(node.location.name, node.name) }} "DIM" ON
            {% for col in source.columns if col.isBusinessKey -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                {{ get_source_transform(col) }} = "DIM"."{{ col.name }}"
            {% endfor %}
            WHERE "DIM"."{{ get_value_by_column_attribute("isSystemCurrentFlag") }}" = 'Y'
            AND (
            {% for col in source.columns if (col.isChangeTracking) -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                {{ get_source_transform(col) }} = "DIM"."{{ col.name }}"
            {% endfor %} )
            {% for col in source.columns if not (   col.isBusinessKey or
                                                    col.isChangeTracking or
                                                    col.isSurrogateKey or
                                                    col.isSystemVersion or
                                                    col.isSystemCurrentFlag or
                                                    col.isSystemStartDate or
                                                    col.isSystemEndDate or
                                                    col.isSystemUpdateDate or
                                                    col.isSystemCreateDate) -%}
                {% if loop.first %}
                    AND (
                {% endif %}
                {% if not loop.first %}
                    OR
                {% endif %}
                NVL( CAST({{ get_source_transform(col) }} as STRING), '**NULL**') <> NVL( CAST("DIM"."{{ col.name }}" as STRING), '**NULL**')
                {% if loop.last %}
                    )
                {% endif %}
            {% endfor %}
        ) AS "SRC"
        ON
        {% for col in source.columns if col.isBusinessKey -%}
            {% if not loop.first %}
                AND
            {% endif %}
            "TGT"."{{ col.name }}" = "SRC"."{{ col.name }}"
        {% endfor %}
        AND "TGT"."{{ get_value_by_column_attribute("isSystemVersion") }}" = "SRC"."{{ get_value_by_column_attribute("isSystemVersion") }}"
        WHEN MATCHED THEN UPDATE SET
        {%- for col in source.columns if not (col.isBusinessKey or col.isSurrogateKey or col.isSystemCreateDate) %}
            "TGT"."{{ col.name }}" = "SRC"."{{ col.name }}"
            {% if not loop.last %}, {% endif %}
        {% endfor -%}
        WHEN NOT MATCHED THEN INSERT (
        {%- for col in source.columns if not col.isSurrogateKey %}
            "{{ col.name }}"
            {% if not loop.last %}, {% endif %}
        {% endfor -%}
        )
        VALUES (
        {%- for col in source.columns if not col.isSurrogateKey %}
            "SRC"."{{ col.name }}"
            {% if not loop.last %}, {% endif %}
        {% endfor -%}
        )

    {% endfor %}

    {% elif has_business_key and not is_type_2 %}
        {% for source in sources %}
            {{ stage('MERGE ' + source.name | string ) }}
            MERGE INTO {{ ref_no_link(node.location.name, node.name) }} "TGT"
            USING (
                SELECT
                {% for col in source.columns if not col.isSurrogateKey %}
                    {% if col.isSystemVersion %}
                    \t1
                    {% elif col.isSystemCurrentFlag %}
                    \t'Y'
                    {% else %}
                        {{ get_source_transform(col) }}
                    {% endif %}
                    AS "{{ col.name }}"
                    {%- if not loop.last %}, {% endif %}
                {% endfor %}
                {{ source.join }})
                AS "SRC"
            ON
            {% for col in source.columns if col.isBusinessKey -%}
                {% if not loop.first %}
                    AND
                {% endif %}
                "SRC"."{{ col.name }}" = "TGT"."{{ col.name }}"
            {% endfor %}
            WHEN MATCHED
            {% for col in source.columns if not (   col.isBusinessKey or
                                                    col.isSurrogateKey or
                                                    col.isSystemVersion or
                                                    col.isSystemCurrentFlag or
                                                    col.isSystemStartDate or
                                                    col.isSystemEndDate or
                                                    col.isSystemUpdateDate or
                                                    col.isSystemCreateDate) %}
                {% if loop.first %}
                    AND (
                {% else %}
                    OR
                {% endif %}
                NVL( CAST("SRC"."{{ col.name }}" as STRING), '**NULL**') <> NVL( CAST("TGT"."{{ col.name }}" as STRING), '**NULL**')
                {% if loop.last %}
                    )
                {% endif %}
            {% endfor %}
            THEN UPDATE SET
            {%- for col in source.columns if not (  col.isBusinessKey or
                                                    col.isSurrogateKey or
                                                    col.isSystemVersion or
                                                    col.isSystemCurrentFlag or
                                                    col.isSystemStartDate or
                                                    col.isSystemEndDate or
                                                    col.isSystemCreateDate) %}
                    "TGT"."{{ col.name }}" = "SRC"."{{ col.name }}"
                {% if not loop.last %}, {% endif %}
            {% endfor %}
            WHEN NOT MATCHED THEN
            INSERT (
            {%- for col in source.columns if not col.isSurrogateKey %}
                "{{ col.name }}"
                {% if not loop.last %}, {% endif %}
            {% endfor -%}
            )
            VALUES (
            {%- for col in source.columns if not col.isSurrogateKey %}
                "SRC"."{{ col.name }}"
                {% if not loop.last %}, {% endif %}
            {% endfor -%}
            )
        {% endfor %}
    {% else %}
        {% for source in sources %}
            {{ stage('Insert ' + source.name | string ) }}
            INSERT INTO {{ ref_no_link(node.location.name, node.name) }}
            (
                {% for col in source.columns if not col.isSurrogateKey %}
                    "{{ col.name }}"
                    {%- if not loop.last -%},{% endif %}
                {% endfor %}
            )

            SELECT
            {% for col in source.columns if not col.isSurrogateKey %}

                {% if col.isSystemVersion %}
                    1
                {% elif col.isSystemCurrentFlag %}
                    'Y'
                {% else %}
                    {{ get_source_transform(col) }}
                {% endif %}
                AS "{{ col.name }}"
                {%- if not loop.last -%}, {% endif %}
                
            {% endfor %}
            {{ source.join }}
        {% endfor %}            
    {% endif %}
	
	{% if config.postSQL %}
		{{ stage('Post-SQL') }}
		{{ config.postSQL }}
	{% endif %}
	
{% endif %}

{% if config.testsEnabled %}
	{% for test in node.tests %}
		{% if test.runOrder == 'After' %}
			{{ test_stage(test.name, test.continueOnFailure) }}
			{{ test.templateString }}
        {% endif %}
	{% endfor %}

	{% for column in columns %}
		{% for test in column.tests %}
			{{ test_stage(column.name + ": " + test.name) }}
			{{ test.templateString }}
		{% endfor %}
	{% endfor %}
{% endif %}` }
            },
            error: null,
        },
    },
    [Runner.ESQLType.fact]: {
        name: "Fact",
        version: 1,
        isDisabled: false,
        metadata: {
            defaultStorageLocation: null,
            nodeMetadataSpec: `capitalized: Fact
plural: Facts
short: FCT
tagColor: '#D9A438'

config:
- groupName: Options
  items:
  - type: materializationSelector
    isRequired: true
    options:
    - table
    - view
  
  - type: businessKeyColumns
    isRequired: false

  - displayName: Enable Tests
    attributeName: testsEnabled
    type: toggleButton
    default: true
    
  - displayName: Pre-SQL
    attributeName: preSQL
    type: textBox
    syntax: sql
    isRequired: false

  - displayName: Post-SQL
    attributeName: postSQL
    type: textBox
    syntax: sql
    isRequired: false

systemColumns:
- displayName: SYSTEM_CREATE_DATE
  transform: CAST(CURRENT_TIMESTAMP AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemCreateDate

- displayName: SYSTEM_UPDATE_DATE
  transform: CAST(CURRENT_TIMESTAMP AS TIMESTAMP)
  dataType: TIMESTAMP
  placement: end
  attributeName: isSystemUpdateDate
`,
            templates: {
                [RunnerOperations.OperationTypes.create]: { templateString: `
    {% if node.materializationType == 'table' %}
        {{ stage('Create Fact Table') }}
    
        CREATE OR REPLACE TABLE {{ ref_no_link(node.location.name, node.name) }}
        (
            {% for col in columns %}
                "{{ col.name }}" {{ col.dataType }}
                {%- if not col.nullable %} NOT NULL
                    {%- if col.defaultValue | length > 0 %} DEFAULT {{ col.defaultValue }}{% endif %}
                {% endif %}
                {%- if col.description | length > 0 %} COMMENT '{{ col.description }}'{% endif %}
                {%- if not loop.last -%}, {% endif %}
            {% endfor %}
        )
        {%- if node.description | length > 0 %} COMMENT = '{{ node.description }}'{% endif %}
    
    
    {% elif node.materializationType == 'view' %}
        {{ stage('Create Fact View') }}
    
        CREATE OR REPLACE VIEW {{ ref_no_link(node.location.name, node.name) }}
        (
            {% for col in columns %}
                "{{ col.name }}"
                {%- if col.description | length > 0 %} COMMENT '{{ col.description }}'{% endif %}
                {%- if not loop.last -%},{% endif %}
            {% endfor %}
        )
        {%- if node.description | length > 0 %} COMMENT = '{{ node.description }}'{% endif %}
        AS
        {% for source in sources %}
    
            {% if loop.first %}SELECT {% endif %}
    
            {% for col in source.columns %}
                {{ get_source_transform(col) }} AS "{{ col.name }}"
                {%- if not loop.last -%}, {% endif %}
            {% endfor %}
            {{ source.join }}
    
            {% if not loop.last %} UNION ALL {% endif %}
        {% endfor %}
    
    {% endif %}
    `
                },
                [RunnerOperations.OperationTypes.run]: { templateString: `
    {% for test in node.tests if config.testsEnabled %}
        {% if test.runOrder == 'Before' %}
            {{ test_stage(test.name, test.continueOnFailure) }}
            {{ test.templateString }}
        {% endif %}
    {% endfor %}

    {% if node.materializationType == 'table' %}
        {% if config.preSQL %}
            {{ stage('Pre-SQL') }}
            {{ config.preSQL }}
        {% endif %}
        
        {% set has_business_key = columns | selectattr("isBusinessKey") | list | length > 0 %}
        
        {% for source in sources %}
        
            {% if has_business_key %}
        
                {{ stage('MERGE ' + source.name | string ) }}
                MERGE INTO {{ ref_no_link(node.location.name, node.name) }} "TGT"
                USING (
                    SELECT
                    {% for col in source.columns %}
                        {{ get_source_transform(col) }} AS "{{ col.name }}"
                        {%- if not loop.last %}, {% endif %}
                    {% endfor %}
                    {{ source.join }})
                    AS "SRC"
                ON
                {% for col in source.columns if col.isBusinessKey -%}
                    {% if not loop.first %}
                        AND
                    {% endif %}
                    "SRC"."{{ col.name }}" = "TGT"."{{ col.name }}"
                {% endfor %}
                WHEN MATCHED
                {% for col in source.columns if not (   col.isBusinessKey or
                                                        col.isSystemUpdateDate or
                                                        col.isSystemCreateDate) %}
                    {% if loop.first %}
                        AND (
                    {% else %}
                        OR
                    {% endif %}
                    NVL( CAST("SRC"."{{ col.name }}" as STRING), '**NULL**') <> NVL( CAST("TGT"."{{ col.name }}" as STRING), '**NULL**')
                    {% if loop.last %}
                        )
                    {% endif %}
                {% endfor %}
                THEN UPDATE SET
                {%- for col in source.columns if not (col.isBusinessKey or col.isSystemCreateDate) %}
                        "TGT"."{{ col.name }}" = "SRC"."{{ col.name }}"
                    {% if not loop.last %}, {% endif %}
                {% endfor %}
                WHEN NOT MATCHED THEN
                INSERT (
                {%- for col in source.columns if not col.isSurrogateKey %}
                    "{{ col.name }}"
                    {% if not loop.last %}, {% endif %}
                {% endfor -%}
                )
                VALUES (
                {%- for col in source.columns if not col.isSurrogateKey %}
                    "SRC"."{{ col.name }}"
                    {% if not loop.last %}, {% endif %}
                {% endfor -%}
                )
        
            {% else %}
        
                {{ stage('Insert ' + source.name | string ) }}
        
                    INSERT INTO {{ ref_no_link(node.location.name, node.name) }}
                    (
                        {% for col in source.columns %}
                            "{{ col.name }}"
                            {%- if not loop.last -%},{% endif %}
                        {% endfor %}
                    )
        
                    SELECT
                    {% for col in source.columns %}
                        {{ get_source_transform(col) }} AS "{{ col.name }}"
                        {%- if not loop.last -%}, {% endif %}
                    {% endfor %}
                    {{ source.join }}
            {% endif %}
        {% endfor %}
        {% if config.postSQL %}
            {{ stage('Post-SQL') }}
            {{ config.postSQL }}
        {% endif %}
    {% endif %}

    {% if config.testsEnabled %}
        {% for test in node.tests %}
            {% if test.runOrder == 'After' %}
                {{ test_stage(test.name, test.continueOnFailure) }}
                {{ test.templateString }}
            {% endif %}
        {% endfor %}

        {% for column in columns %}
            {% for test in column.tests %}
                {{ test_stage(column.name + ": " + test.name) }}
                {{ test.templateString }}
            {% endfor %}
        {% endfor %}
    {% endif %}
    
    ` },
            },
            error: null
        }
    }
});
exports.getDefaultStepTypes = getDefaultStepTypes;
//# sourceMappingURL=StepTypes.js.map