"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GetEnabledColumnTestIDs = exports.GetAppliedColumnTestsDefinitions = exports.GetAppliedColumnTestsSafe = exports.GetEnabledColumnTestDefinitions = exports.ToggleAppliedColumnTestForColumn = exports.UpdateEnabledColumnTestsForStep = exports.IsFailedTestQueryResult = exports.IsTestQueryResult = exports.EditNodeTestAtIndex = exports.DeleteNodeTestAtIndex = exports.CreateAndAddFreshNodeTestToNode = exports.UpdateOrderOfNodeTests = exports.GetAppliedNodeTestsSafe = exports.DefaultColumnTests = exports.RNodeTest = exports.RAcceptedValues = exports.EDefaultColumnTestIDs = exports.ENodeTestProperties = exports.ETestRunOrder = exports.ETestTypes = exports.ETestTabs = void 0;
const immer = __importStar(require("immer"));
const CommonOperations = __importStar(require("./CommonOperations"));
const StepOperations = __importStar(require("./StepOperations"));
const MetadataOperations = __importStar(require("./MetadataOperations"));
const Logging = __importStar(require("./Logging"));
const Common = __importStar(require("./Common"));
const R = __importStar(require("runtypes"));
const TestLogger = Logging.GetLogger(Logging.LoggingArea.Testing);
// Enums, vars
var ETestTabs;
(function (ETestTabs) {
    ETestTabs["node"] = "node";
    ETestTabs["column"] = "column";
})(ETestTabs = exports.ETestTabs || (exports.ETestTabs = {}));
var ETestTypes;
(function (ETestTypes) {
    ETestTypes["nodeLevelTest"] = "nodeLevelTest";
    ETestTypes["columnLevelTest"] = "columnLevelTest";
})(ETestTypes = exports.ETestTypes || (exports.ETestTypes = {}));
var ETestRunOrder;
(function (ETestRunOrder) {
    ETestRunOrder["before"] = "Before";
    ETestRunOrder["after"] = "After";
})(ETestRunOrder = exports.ETestRunOrder || (exports.ETestRunOrder = {}));
var ENodeTestProperties;
(function (ENodeTestProperties) {
    ENodeTestProperties["name"] = "name";
    ENodeTestProperties["description"] = "description";
    ENodeTestProperties["continueOnFailure"] = "continueOnFailure";
    ENodeTestProperties["runOrder"] = "runOrder";
    ENodeTestProperties["templateString"] = "templateString";
})(ENodeTestProperties = exports.ENodeTestProperties || (exports.ENodeTestProperties = {}));
var EDefaultColumnTestIDs;
(function (EDefaultColumnTestIDs) {
    EDefaultColumnTestIDs["hasNull"] = "hasNull";
    EDefaultColumnTestIDs["isDistinct"] = "isDistinct";
    EDefaultColumnTestIDs["acceptedValues"] = "acceptedValues";
})(EDefaultColumnTestIDs = exports.EDefaultColumnTestIDs || (exports.EDefaultColumnTestIDs = {}));
exports.RAcceptedValues = R.Record({
    strictMatch: R.Boolean,
    values: R.Array(R.String)
});
exports.RNodeTest = R.Record({
    name: R.String,
    description: R.String,
    continueOnFailure: R.Boolean,
    runOrder: R.Union(R.Literal("Before"), R.Literal("After")),
    templateString: R.String
});
// Dictionaries
exports.DefaultColumnTests = {
    hasNull: {
        name: "Null",
        description: "checks for any null values",
        getTemplateString: (col) => `
            SELECT * FROM {{ ref(node.location.name, node.name)}} 
            WHERE "${col.name}" IS NULL`,
        columnTestDefinitionID: EDefaultColumnTestIDs.hasNull,
        continueOnFailure: true,
        attributes: {},
        runOrder: ETestRunOrder.after
    },
    isDistinct: {
        name: "Unique",
        description: "check if value is unique",
        getTemplateString: (col) => `
            SELECT "${col.name}", COUNT(*)
            FROM {{ ref(node.location.name, node.name)}} 
            GROUP BY "${col.name}"
            HAVING COUNT(*) > 1`,
        columnTestDefinitionID: EDefaultColumnTestIDs.isDistinct,
        continueOnFailure: true,
        attributes: {},
        runOrder: ETestRunOrder.after
    }
};
/**
 * /////////////////////////
 * /// Node and Column Test Operations
 * /////////////////////////
 */
const UpdateTestingMetadataForStepFS = (teamID, workspaceID, stepCounter, updatedTestingPayload) => {
    const stepsDocRef = CommonOperations.ECollectionRefs.steps;
    const updates = {
        metadata: Object.assign({}, updatedTestingPayload)
    };
    /**
     * FS method set() will throw an exception before returning a promise if fed invalid args.
     * To account for this, we create a new promise with a try/catch block
     *
     * We handle the happy path of set() returning a promise inside of the try block (and handling exceptions as we normally would in the .catch() chanining).
     *
     * The catch block will catch the synchronous return in case of invalid args and reject, which will allow us to actually catch and handle the error on the front end.
     *
     * This implementation will likely be refactored after addressing ticket CD-2068
     */
    return new Promise((resolve, reject) => {
        try {
            StepOperations.GetStepDocumentFS(teamID, workspaceID, stepsDocRef, stepCounter).set(updates, { merge: true })
                .then(resolve)
                .catch(reject);
        }
        catch (error) {
            reject(error);
            TestLogger.alertContext({ orgID: teamID, envID: workspaceID }, `SQL Testing threw an error inside of step: ${stepCounter}`, error); // in the case of invalid arguments, we need to log
        }
    });
};
/**
 * /////////////////////////
 * /// Node Test Operations
 * /////////////////////////
 */
const GetAppliedNodeTestsSafe = (sqlOperation) => {
    return Common.getValueSafe(sqlOperation, ["metadata", "appliedNodeTests"], []);
};
exports.GetAppliedNodeTestsSafe = GetAppliedNodeTestsSafe;
const UpdateOrderOfNodeTests = (teamID, workspaceID, stepCounter, reorderedNodeTests) => {
    return UpdateTestingMetadataForStepFS(teamID, workspaceID, stepCounter, { appliedNodeTests: reorderedNodeTests });
};
exports.UpdateOrderOfNodeTests = UpdateOrderOfNodeTests;
const CreateAndAddFreshNodeTestToNode = (teamID, workspaceID, stepCounter, currentNodeTests) => {
    const freshNodeTest = InitializeNewNodeTest(currentNodeTests);
    return AddNodeTestToNode(teamID, workspaceID, stepCounter, freshNodeTest, currentNodeTests);
};
exports.CreateAndAddFreshNodeTestToNode = CreateAndAddFreshNodeTestToNode;
const AddNodeTestToNode = (teamID, workspaceID, stepCounter, newOrUpdatedTest, currentNodeTests) => {
    const updatedNodeTests = [...currentNodeTests, newOrUpdatedTest];
    return UpdateTestingMetadataForStepFS(teamID, workspaceID, stepCounter, { appliedNodeTests: updatedNodeTests });
};
const DeleteNodeTestAtIndex = (teamID, workspaceID, stepCounter, testIndex, currentNodeTests) => {
    const updatedNodeTests = immer.produce(currentNodeTests, (draftState) => {
        draftState.splice(testIndex, 1);
    });
    return UpdateTestingMetadataForStepFS(teamID, workspaceID, stepCounter, { appliedNodeTests: updatedNodeTests });
};
exports.DeleteNodeTestAtIndex = DeleteNodeTestAtIndex;
const EditNodeTestAtIndex = (teamID, workspaceID, stepCounter, testIndex, currentNodeTests, editedNodeTest) => {
    const updatedNodeTests = immer.produce(currentNodeTests, (draftState) => {
        draftState.splice(testIndex, 1, editedNodeTest);
    });
    const isNameUnique = IsNodeTestNameUnique(editedNodeTest, currentNodeTests, testIndex); // we use the node test name as the unique identifier, so this is important
    if (!isNameUnique)
        return Common.ThrowErrorAsPromise(new Error("Rename failed: Test name must be unique"));
    if (editedNodeTest.name.trim() === "")
        return Common.ThrowErrorAsPromise(new Error("Rename failed: Test name cannot be an empty string or only whitespace"));
    return UpdateTestingMetadataForStepFS(teamID, workspaceID, stepCounter, { appliedNodeTests: updatedNodeTests });
};
exports.EditNodeTestAtIndex = EditNodeTestAtIndex;
const InitializeNewNodeTest = (currentNodeTests) => {
    let counter = 1;
    let newTest = {
        name: "Test",
        description: "",
        templateString: "",
        continueOnFailure: true,
        runOrder: ETestRunOrder.after
    };
    const getTestWithUniqueName = (test) => {
        const isNameUnique = IsNodeTestNameUnique(test, currentNodeTests, -1); // -1 should never match the index of any of the currentNodeTests
        if (isNameUnique)
            return test;
        test.name = `Test ${counter}`;
        counter++;
        return getTestWithUniqueName(test);
    };
    return getTestWithUniqueName(newTest);
};
const IsNodeTestNameUnique = (nodeTestToCheck, currentNodeTests, testIndex) => {
    let isUnique = true;
    currentNodeTests.forEach((test, index) => {
        if (test.name === nodeTestToCheck.name.trim() && index !== testIndex)
            isUnique = false;
    });
    return isUnique;
};
const IsTestQueryResult = (queryResult) => {
    return queryResult.type.toLowerCase().includes("test");
};
exports.IsTestQueryResult = IsTestQueryResult;
const IsFailedTestQueryResult = (queryResult) => {
    Common.assert(Logging.LoggingArea.Runner, (0, exports.IsTestQueryResult)(queryResult), `Query result type is invalid: ${queryResult.type}`);
    return queryResult.rows.length > 0;
};
exports.IsFailedTestQueryResult = IsFailedTestQueryResult;
/**
 * ///////////////////////////////////////////////////////////////////////
 * /////////////////// Column Test Operations ////////////////////////////
 * ///////////////////////////////////////////////////////////////////////
 *
 *
 *
 * Note the difference between "enabled" and "applied" column tests
 *
 * enabled column test ids are scoped to the node level (string)
 *   - these are tests ids that have been selected in the dropdown selector in the col testing menu
 *   - corresponding test definitions are made available to each column of the node to apply to the column as an applied column test
 *
 * applied column tests are scoped to an individual column of the node (IAppliedColumnTest)
 *   - once a test is enabled as above, then the column may apply that enabled test (or unapply it)
 *   - whether or not test applied indicated by boolean
 *
 * a column test definition is an IColumnTest
 *
 * ////////////////////////////////////////////////////////
 */
const UpdateEnabledColumnTestsForStep = (teamID, workspaceID, stepCounter, updatedEnabledColumnTestIDs, stepMetadata) => {
    const intializeAndCleanupAppliedColumnTests = (column) => {
        column.appliedColumnTests = (0, exports.GetAppliedColumnTestsSafe)(column);
        // initialize new appliedColTest set to false (not selected)
        updatedEnabledColumnTestIDs.forEach(enabledTestID => {
            if (!(enabledTestID in column.appliedColumnTests)) {
                column.appliedColumnTests[enabledTestID] = false;
            }
        });
        // cleanup any applied tests for col that are not included in enabledColTestIDs
        for (const appliedTestID in column.appliedColumnTests) {
            if (!updatedEnabledColumnTestIDs.includes(appliedTestID)) {
                delete column.appliedColumnTests[appliedTestID];
            }
        }
    };
    const updatedColumns = MetadataOperations.EditAllColumns(stepMetadata.columns, intializeAndCleanupAppliedColumnTests);
    return UpdateTestingMetadataForStepFS(teamID, workspaceID, stepCounter, { columns: updatedColumns, enabledColumnTestIDs: updatedEnabledColumnTestIDs });
};
exports.UpdateEnabledColumnTestsForStep = UpdateEnabledColumnTestsForStep;
const ToggleAppliedColumnTestForColumn = (teamID, workspaceID, stepCounter, stepColumns, toggledColumnTestID, stepColumnIndex) => {
    const updateToggledColumns = (column) => {
        column.appliedColumnTests[toggledColumnTestID] = !column.appliedColumnTests[toggledColumnTestID];
    };
    const updatedColumns = MetadataOperations.EditColumnByColumnIndex(stepColumns, stepColumnIndex, updateToggledColumns);
    return UpdateTestingMetadataForStepFS(teamID, workspaceID, stepCounter, { columns: updatedColumns });
};
exports.ToggleAppliedColumnTestForColumn = ToggleAppliedColumnTestForColumn;
const GetEnabledColumnTestDefinitions = (stepData) => {
    const enabledColumnTestIDs = Common.getValueSafe(stepData, ["operation", "metadata", "enabledColumnTestIDs"], []);
    return Object.values(exports.DefaultColumnTests).filter(colTest => enabledColumnTestIDs.includes(colTest.columnTestDefinitionID));
};
exports.GetEnabledColumnTestDefinitions = GetEnabledColumnTestDefinitions;
const GetAppliedColumnTestsSafe = (columnMetadata) => {
    return Common.getValueSafe(columnMetadata, ["appliedColumnTests"], {});
};
exports.GetAppliedColumnTestsSafe = GetAppliedColumnTestsSafe;
const GetAppliedColumnTestsDefinitions = (columnMetadata) => {
    const appliedColumnTests = (0, exports.GetAppliedColumnTestsSafe)(columnMetadata);
    return Object.values(exports.DefaultColumnTests).filter(colTest => {
        return Object.keys(appliedColumnTests).includes(colTest.columnTestDefinitionID) && appliedColumnTests[colTest.columnTestDefinitionID] === true;
    });
};
exports.GetAppliedColumnTestsDefinitions = GetAppliedColumnTestsDefinitions;
const GetEnabledColumnTestIDs = (stepOperation) => {
    return Common.getValueSafe(stepOperation, ["metadata", "enabledColumnTestIDs"], []);
};
exports.GetEnabledColumnTestIDs = GetEnabledColumnTestIDs;
//# sourceMappingURL=TestingOperations.js.map